# Licensed under a 3-clause BSD style license - see LICENSE.rst
# pylint: disable=invalid-name
import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_array_equal

from astropy import units as u
from astropy.modeling.fitting import (
    DogBoxLSQFitter,
    LevMarLSQFitter,
    LMLSQFitter,
    TRFLSQFitter,
)
from astropy.modeling.models import (
    Gaussian1D,
    Identity,
    Mapping,
    Rotation2D,
    Shift,
    UnitsMapping,
)
from astropy.utils import NumpyRNGContext
from astropy.utils.compat.optional_deps import HAS_SCIPY

fitters = [LevMarLSQFitter, TRFLSQFitter, LMLSQFitter, DogBoxLSQFitter]
fitters_bounds = [LevMarLSQFitter, TRFLSQFitter, DogBoxLSQFitter]


def test_swap_axes():
    x = np.zeros((2, 3))
    y = np.ones((2, 3))

    mapping = Mapping((1, 0))
    assert mapping(1, 2) == (2.0, 1.0)
    assert mapping.inverse(2, 1) == (1, 2)
    assert_array_equal(mapping(x, y), (y, x))
    assert_array_equal(mapping.inverse(y, x), (x, y))


def test_duplicate_axes():
    mapping = Mapping((0, 1, 0, 1))
    assert mapping(1, 2) == (1.0, 2.0, 1.0, 2)
    assert mapping.inverse(1, 2, 1, 2) == (1, 2)
    assert mapping.inverse.n_inputs == 4
    assert mapping.inverse.n_outputs == 2


def test_drop_axes_1():
    mapping = Mapping((0,), n_inputs=2)
    assert mapping(1, 2) == (1.0)


def test_drop_axes_2():
    mapping = Mapping((1,))
    assert mapping(1, 2) == (2.0)
    MESSAGE = (
        r"Mappings such as .* that drop one or more of their inputs are not invertible"
        r" at this time"
    )
    with pytest.raises(NotImplementedError, match=MESSAGE):
        mapping.inverse


def test_drop_axes_3():
    mapping = Mapping((1,), n_inputs=2)
    assert mapping.n_inputs == 2
    rotation = Rotation2D(60)
    model = rotation | mapping
    assert_allclose(model(1, 2), 1.86602540378)


@pytest.mark.parametrize("name", [None, "test_name"])
def test_bad_inputs(name):
    mapping = Mapping((1, 0), name=name)

    if name is None:
        name = "Mapping"

    x = [np.ones((2, 3)) * idx for idx in range(5)]
    for idx in range(1, 6):
        if idx == 2:
            continue

        MESSAGE = f"{name} expects 2 inputs; got {idx}"
        with pytest.raises(TypeError, match=MESSAGE):
            mapping.evaluate(*x[:idx])


def test_identity():
    x = np.zeros((2, 3))
    y = np.ones((2, 3))

    ident1 = Identity(1)
    shift = Shift(1)
    rotation = Rotation2D(angle=60)
    model = ident1 & shift | rotation
    assert_allclose(model(1, 2), (-2.098076211353316, 2.3660254037844393))
    res_x, res_y = model(x, y)
    assert_allclose(
        (res_x, res_y),
        (
            np.array(
                [
                    [-1.73205081, -1.73205081, -1.73205081],
                    [-1.73205081, -1.73205081, -1.73205081],
                ]
            ),
            np.array(
                [
                    [1.0, 1.0, 1.0],
                    [1.0, 1.0, 1.0],
                ]
            ),
        ),
    )
    assert_allclose(model.inverse(res_x, res_y), (x, y), atol=1.0e-10)


# https://github.com/astropy/astropy/pull/6018
@pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
@pytest.mark.parametrize("fitter", fitters_bounds)
def test_fittable_compound(fitter):
    fitter = fitter()

    m = Identity(1) | Mapping((0,)) | Gaussian1D(1, 5, 4)
    x = np.arange(10)
    y_real = m(x)
    dy = 0.005
    with NumpyRNGContext(1234567):
        n = np.random.normal(0.0, dy, x.shape)
    y_noisy = y_real + n
    new_model = fitter(m, x, y_noisy)
    y_fit = new_model(x)
    assert_allclose(y_fit, y_real, atol=dy)


def test_identity_repr():
    m = Identity(1, name="foo")
    assert repr(m) == "<Identity(1, name='foo')>"

    m = Identity(1)
    assert repr(m) == "<Identity(1)>"


def test_mapping_repr():
    m = Mapping([0, 1], name="foo")
    assert repr(m) == "<Mapping([0, 1], name='foo')>"

    m = Mapping([0, 1])
    assert repr(m) == "<Mapping([0, 1])>"


class TestUnitsMapping:
    def test___init__(self):
        # Set values
        model = UnitsMapping(
            ((u.m, None),),
            input_units_equivalencies="test_eqiv",
            input_units_allow_dimensionless=True,
            name="test",
        )
        assert model._mapping == ((u.m, None),)
        assert model._input_units_strict == {"x": True}
        assert model.input_units_equivalencies == "test_eqiv"
        assert model.input_units_allow_dimensionless == {"x": True}
        assert model.name == "test"
        assert model._input_units == {"x": u.m}

        # Default values
        model = UnitsMapping(((u.K, None),))
        assert model._mapping == ((u.K, None),)
        assert model._input_units_strict == {"x": True}
        assert model.input_units_equivalencies is None
        assert model.input_units_allow_dimensionless == {"x": False}
        assert model.name is None
        assert model._input_units == {"x": u.K}

        # Error
        MESSAGE = r"If one return unit is None, then all must be None"
        with pytest.raises(ValueError, match=MESSAGE):
            UnitsMapping(((u.m, None), (u.m, u.K)))

    def test_evaluate(self):
        model = UnitsMapping(((u.m, None),))
        assert model(10 * u.m) == 10

        model = UnitsMapping(((u.m, u.K),))
        assert model(10 * u.m) == 10 * u.K

        model = UnitsMapping(
            ((u.m, None), (u.K, None)),
        )
        assert model(10 * u.m, 20 * u.K) == (10, 20)

        model = UnitsMapping(
            ((u.m, u.K), (u.K, u.m)),
        )
        assert model(10 * u.m, 20 * u.K) == (10 * u.K, 20 * u.m)

    def test_repr(self):
        model = UnitsMapping(((u.m, None),), name="foo")
        assert repr(model) == f"<UnitsMapping((({repr(u.m)}, None),), name='foo')>"

        model = UnitsMapping(((u.m, None),))
        assert repr(model) == f"<UnitsMapping((({repr(u.m)}, None),))>"



# Licensed under a 3-clause BSD style license - see PYFITS.rst

"""
A package for reading and writing FITS files and manipulating their
contents.

A module for reading and writing Flexible Image Transport System
(FITS) files.  This file format was endorsed by the International
Astronomical Union in 1999 and mandated by NASA as the standard format
for storing high energy astrophysics data.  For details of the FITS
standard, see the NASA/Science Office of Standards and Technology
publication, NOST 100-2.0.
"""

from astropy import config as _config

# Set module-global boolean variables
# TODO: Make it possible to set these variables via environment variables
# again, once support for that is added to Astropy


class Conf(_config.ConfigNamespace):
    """
    Configuration parameters for `astropy.io.fits`.
    """

    enable_record_valued_keyword_cards = _config.ConfigItem(
        True,
        "If True, enable support for record-valued keywords as described by "
        "FITS WCS distortion paper. Otherwise they are treated as normal "
        "keywords.",
        aliases=["astropy.io.fits.enabled_record_valued_keyword_cards"],
    )
    extension_name_case_sensitive = _config.ConfigItem(
        False,
        "If True, extension names (i.e. the ``EXTNAME`` keyword) should be "
        "treated as case-sensitive.",
    )
    strip_header_whitespace = _config.ConfigItem(
        True,
        "If True, automatically remove trailing whitespace for string values in"
        " headers. Otherwise the values are returned verbatim, with all "
        "whitespace intact.",
    )
    use_memmap = _config.ConfigItem(
        True,
        "If True, use memory-mapped file access to read/write the data in "
        "FITS files. This generally provides better performance, especially "
        "for large files, but may affect performance in I/O-heavy "
        "applications.",
    )
    lazy_load_hdus = _config.ConfigItem(
        True,
        "If True, use lazy loading of HDUs when opening FITS files by "
        "default; that is fits.open() will only seek for and read HDUs on "
        "demand rather than reading all HDUs at once.  See the documentation "
        "for fits.open() for more details.",
    )
    enable_uint = _config.ConfigItem(
        True,
        "If True, default to recognizing the convention for representing "
        "unsigned integers in FITS--if an array has BITPIX > 0, BSCALE = 1, "
        "and BZERO = 2**BITPIX, represent the data as unsigned integers "
        "per this convention.",
    )


conf = Conf()


# Public API compatibility imports
# These need to come after the global config variables, as some of the
# submodules use them
from . import card, column, convenience, hdu
from .card import *
from .column import *
from .convenience import *
from .diff import *
from .fitsrec import FITS_rec, FITS_record
from .hdu import *
from .hdu.compressed import CompImageSection
from .hdu.groups import GroupData
from .hdu.hdulist import fitsopen as open
from .hdu.image import Section
from .header import Header
from .verify import VerifyError

__all__ = (
    ["Conf", "conf"]
    + card.__all__
    + column.__all__
    + convenience.__all__
    + hdu.__all__
    + [
        "FITS_record",
        "FITS_rec",
        "GroupData",
        "open",
        "Section",
        "Header",
        "VerifyError",
        "conf",
    ]
)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This subpackage contains classes and functions for celestial coordinates
of astronomical objects. It also contains a framework for conversions
between coordinate systems.
"""

from .core import *
from .errors import *
from .utils import *



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""This module implements the base CCDData class."""

import itertools

import numpy as np

from astropy import log
from astropy import units as u
from astropy.io import fits, registry
from astropy.utils.decorators import sharedmethod
from astropy.wcs import WCS

from .compat import NDDataArray
from .nduncertainty import (
    InverseVariance,
    NDUncertainty,
    StdDevUncertainty,
    VarianceUncertainty,
)

__all__ = ["CCDData", "fits_ccddata_reader", "fits_ccddata_writer"]

_known_uncertainties = (StdDevUncertainty, VarianceUncertainty, InverseVariance)
_unc_name_to_cls = {cls.__name__: cls for cls in _known_uncertainties}
_unc_cls_to_name = {cls: cls.__name__ for cls in _known_uncertainties}

# Global value which can turn on/off the unit requirements when creating a
# CCDData. Should be used with care because several functions actually break
# if the unit is None!
_config_ccd_requires_unit = True


def _arithmetic(op):
    """Decorator factory which temporarily disables the need for a unit when
    creating a new CCDData instance. The final result must have a unit.

    Parameters
    ----------
    op : function
        The function to apply. Supported are:

        - ``np.add``
        - ``np.subtract``
        - ``np.multiply``
        - ``np.true_divide``

    Notes
    -----
    Should only be used on CCDData ``add``, ``subtract``, ``divide`` or
    ``multiply`` because only these methods from NDArithmeticMixin are
    overwritten.
    """

    def decorator(func):
        def inner(self, operand, operand2=None, **kwargs):
            global _config_ccd_requires_unit
            _config_ccd_requires_unit = False
            result = self._prepare_then_do_arithmetic(op, operand, operand2, **kwargs)
            # Wrap it again as CCDData so it checks the final unit.
            _config_ccd_requires_unit = True
            return result.__class__(result)

        inner.__doc__ = f"See `astropy.nddata.NDArithmeticMixin.{func.__name__}`."
        return sharedmethod(inner)

    return decorator


def _uncertainty_unit_equivalent_to_parent(uncertainty_type, unit, parent_unit):
    if uncertainty_type is StdDevUncertainty:
        return unit == parent_unit
    elif uncertainty_type is VarianceUncertainty:
        return unit == (parent_unit**2)
    elif uncertainty_type is InverseVariance:
        return unit == (1 / (parent_unit**2))
    raise ValueError(f"unsupported uncertainty type: {uncertainty_type}")


class CCDData(NDDataArray):
    """A class describing basic CCD data.

    The CCDData class is based on the NDData object and includes a data array,
    uncertainty frame, mask frame, flag frame, meta data, units, and WCS
    information for a single CCD image.

    Parameters
    ----------
    data : `~astropy.nddata.CCDData`-like or array-like
        The actual data contained in this `~astropy.nddata.CCDData` object.
        Note that the data will always be saved by *reference*, so you should
        make a copy of the ``data`` before passing it in if that's the desired
        behavior.

    uncertainty : `~astropy.nddata.StdDevUncertainty`, \
            `~astropy.nddata.VarianceUncertainty`, \
            `~astropy.nddata.InverseVariance`, `numpy.ndarray` or \
            None, optional
        Uncertainties on the data. If the uncertainty is a `numpy.ndarray`, it
        it assumed to be, and stored as, a `~astropy.nddata.StdDevUncertainty`.
        Default is ``None``.

    mask : `numpy.ndarray` or None, optional
        Mask for the data, given as a boolean Numpy array with a shape
        matching that of the data. The values must be `False` where
        the data is *valid* and `True` when it is not (like Numpy
        masked arrays). If ``data`` is a numpy masked array, providing
        ``mask`` here will causes the mask from the masked array to be
        ignored.
        Default is ``None``.

    flags : `numpy.ndarray` or `~astropy.nddata.FlagCollection` or None, \
            optional
        Flags giving information about each pixel. These can be specified
        either as a Numpy array of any type with a shape matching that of the
        data, or as a `~astropy.nddata.FlagCollection` instance which has a
        shape matching that of the data.
        Default is ``None``.

    wcs : `~astropy.wcs.WCS` or None, optional
        WCS-object containing the world coordinate system for the data.
        Default is ``None``.

    meta : dict-like object or None, optional
        Metadata for this object. "Metadata" here means all information that
        is included with this object but not part of any other attribute
        of this particular object, e.g. creation date, unique identifier,
        simulation parameters, exposure time, telescope name, etc.

    unit : `~astropy.units.Unit` or str, optional
        The units of the data.
        Default is ``None``.

        .. warning::

            If the unit is ``None`` or not otherwise specified it will raise a
            ``ValueError``

    psf : `numpy.ndarray` or None, optional
        Image representation of the PSF at the center of this image. In order
        for convolution to be flux-preserving, this should generally be
        normalized to sum to unity.

    Raises
    ------
    ValueError
        If the ``uncertainty`` or ``mask`` inputs cannot be broadcast (e.g.,
        match shape) onto ``data``.

    Methods
    -------
    read(\\*args, \\**kwargs)
        ``Classmethod`` to create an CCDData instance based on a ``FITS`` file.
        This method uses :func:`fits_ccddata_reader` with the provided
        parameters.
    write(\\*args, \\**kwargs)
        Writes the contents of the CCDData instance into a new ``FITS`` file.
        This method uses :func:`fits_ccddata_writer` with the provided
        parameters.

    Attributes
    ----------
    known_invalid_fits_unit_strings
        A dictionary that maps commonly-used fits unit name strings that are
        technically invalid to the correct valid unit type (or unit string).
        This is primarily for variant names like "ELECTRONS/S" which are not
        formally valid, but are unambiguous and frequently enough encountered
        that it is convenient to map them to the correct unit.

    Notes
    -----
    `~astropy.nddata.CCDData` objects can be easily converted to a regular
     Numpy array using `numpy.asarray`.

    For example::

        >>> from astropy.nddata import CCDData
        >>> import numpy as np
        >>> x = CCDData([1,2,3], unit='adu')
        >>> np.asarray(x)
        array([1, 2, 3])

    This is useful, for example, when plotting a 2D image using
    matplotlib.

        >>> from astropy.nddata import CCDData
        >>> from matplotlib import pyplot as plt   # doctest: +SKIP
        >>> x = CCDData([[1,2,3], [4,5,6]], unit='adu')
        >>> plt.imshow(x)   # doctest: +SKIP

    """

    def __init__(self, *args, **kwd):
        if "meta" not in kwd:
            kwd["meta"] = kwd.pop("header", None)
        if "header" in kwd:
            raise ValueError("can't have both header and meta.")

        super().__init__(*args, **kwd)
        if self._wcs is not None:
            llwcs = self._wcs.low_level_wcs
            if not isinstance(llwcs, WCS):
                raise TypeError("the wcs must be a WCS instance.")
            self._wcs = llwcs

        # Check if a unit is set. This can be temporarily disabled by the
        # _CCDDataUnit contextmanager.
        if _config_ccd_requires_unit and self.unit is None:
            raise ValueError("a unit for CCDData must be specified.")

    def _slice_wcs(self, item):
        """
        Override the WCS slicing behaviour so that the wcs attribute continues
        to be an `astropy.wcs.WCS`.
        """
        if self.wcs is None:
            return None

        try:
            return self.wcs[item]
        except Exception as err:
            self._handle_wcs_slicing_error(err, item)

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        self._data = value

    @property
    def wcs(self):
        return self._wcs

    @wcs.setter
    def wcs(self, value):
        if value is not None and not isinstance(value, WCS):
            raise TypeError("the wcs must be a WCS instance.")
        self._wcs = value

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, value):
        self._unit = u.Unit(value)

    @property
    def psf(self):
        return self._psf

    @psf.setter
    def psf(self, value):
        if value is not None and not isinstance(value, np.ndarray):
            raise TypeError("The psf must be a numpy array.")
        self._psf = value

    @property
    def header(self):
        return self._meta

    @header.setter
    def header(self, value):
        self.meta = value

    @property
    def uncertainty(self):
        return self._uncertainty

    @uncertainty.setter
    def uncertainty(self, value):
        if value is not None:
            if isinstance(value, NDUncertainty):
                if getattr(value, "_parent_nddata", None) is not None:
                    value = value.__class__(value, copy=False)
                self._uncertainty = value
            elif isinstance(value, np.ndarray):
                if value.shape != self.shape:
                    raise ValueError("uncertainty must have same shape as data.")
                self._uncertainty = StdDevUncertainty(value)
                log.info(
                    "array provided for uncertainty; assuming it is a "
                    "StdDevUncertainty."
                )
            else:
                raise TypeError(
                    "uncertainty must be an instance of a "
                    "NDUncertainty object or a numpy array."
                )
            self._uncertainty.parent_nddata = self
        else:
            self._uncertainty = value

    def to_hdu(
        self,
        hdu_mask="MASK",
        hdu_uncertainty="UNCERT",
        hdu_flags=None,
        wcs_relax=True,
        key_uncertainty_type="UTYPE",
        as_image_hdu=False,
        hdu_psf="PSFIMAGE",
    ):
        """Creates an HDUList object from a CCDData object.

        Parameters
        ----------
        hdu_mask, hdu_uncertainty, hdu_flags, hdu_psf : str or None, optional
            If it is a string append this attribute to the HDUList as
            `~astropy.io.fits.ImageHDU` with the string as extension name.
            Flags are not supported at this time. If ``None`` this attribute
            is not appended.
            Default is ``'MASK'`` for mask, ``'UNCERT'`` for uncertainty,
            ``'PSFIMAGE'`` for psf, and `None` for flags.

        wcs_relax : bool
            Value of the ``relax`` parameter to use in converting the WCS to a
            FITS header using `~astropy.wcs.WCS.to_header`. The common
            ``CTYPE`` ``RA---TAN-SIP`` and ``DEC--TAN-SIP`` requires
            ``relax=True`` for the ``-SIP`` part of the ``CTYPE`` to be
            preserved.

        key_uncertainty_type : str, optional
            The header key name for the class name of the uncertainty (if any)
            that is used to store the uncertainty type in the uncertainty hdu.
            Default is ``UTYPE``.

            .. versionadded:: 3.1

        as_image_hdu : bool
            If this option is `True`, the first item of the returned
            `~astropy.io.fits.HDUList` is a `~astropy.io.fits.ImageHDU`, instead
            of the default `~astropy.io.fits.PrimaryHDU`.

        Raises
        ------
        ValueError
            - If ``self.mask`` is set but not a `numpy.ndarray`.
            - If ``self.uncertainty`` is set but not a astropy uncertainty type.
            - If ``self.uncertainty`` is set but has another unit then
              ``self.data``.

        NotImplementedError
            Saving flags is not supported.

        Returns
        -------
        hdulist : `~astropy.io.fits.HDUList`
        """
        if isinstance(self.header, fits.Header):
            # Copy here so that we can modify the HDU header by adding WCS
            # information without changing the header of the CCDData object.
            header = self.header.copy()
        else:
            # Because _insert_in_metadata_fits_safe is written as a method
            # we need to create a dummy CCDData instance to hold the FITS
            # header we are constructing. This probably indicates that
            # _insert_in_metadata_fits_safe should be rewritten in a more
            # sensible way...
            dummy_ccd = CCDData([1], meta=fits.Header(), unit="adu")
            for k, v in self.header.items():
                dummy_ccd._insert_in_metadata_fits_safe(k, v)
            header = dummy_ccd.header
        if self.unit is not u.dimensionless_unscaled:
            header["bunit"] = self.unit.to_string()
        if self.wcs:
            # Simply extending the FITS header with the WCS can lead to
            # duplicates of the WCS keywords; iterating over the WCS
            # header should be safer.
            #
            # Turns out if I had read the io.fits.Header.extend docs more
            # carefully, I would have realized that the keywords exist to
            # avoid duplicates and preserve, as much as possible, the
            # structure of the commentary cards.
            #
            # Note that until astropy/astropy#3967 is closed, the extend
            # will fail if there are comment cards in the WCS header but
            # not header.
            wcs_header = self.wcs.to_header(relax=wcs_relax)
            header.extend(wcs_header, useblanks=False, update=True)

        if as_image_hdu:
            hdus = [fits.ImageHDU(self.data, header)]
        else:
            hdus = [fits.PrimaryHDU(self.data, header)]

        if hdu_mask and self.mask is not None:
            # Always assuming that the mask is a np.ndarray (check that it has
            # a 'shape').
            if not hasattr(self.mask, "shape"):
                raise ValueError("only a numpy.ndarray mask can be saved.")

            # Convert boolean mask to uint since io.fits cannot handle bool.
            hduMask = fits.ImageHDU(self.mask.astype(np.uint8), name=hdu_mask)
            hdus.append(hduMask)

        if hdu_uncertainty and self.uncertainty is not None:
            # We need to save some kind of information which uncertainty was
            # used so that loading the HDUList can infer the uncertainty type.
            # No idea how this can be done so only allow StdDevUncertainty.
            uncertainty_cls = self.uncertainty.__class__
            if uncertainty_cls not in _known_uncertainties:
                raise ValueError(
                    f"only uncertainties of type {_known_uncertainties} can be saved."
                )
            uncertainty_name = _unc_cls_to_name[uncertainty_cls]

            hdr_uncertainty = fits.Header()
            hdr_uncertainty[key_uncertainty_type] = uncertainty_name

            # Assuming uncertainty is an StdDevUncertainty save just the array
            # this might be problematic if the Uncertainty has a unit differing
            # from the data so abort for different units. This is important for
            # astropy > 1.2
            if hasattr(self.uncertainty, "unit") and self.uncertainty.unit is not None:
                if not _uncertainty_unit_equivalent_to_parent(
                    uncertainty_cls, self.uncertainty.unit, self.unit
                ):
                    raise ValueError(
                        "saving uncertainties with a unit that is not "
                        "equivalent to the unit from the data unit is not "
                        "supported."
                    )

            hduUncert = fits.ImageHDU(
                self.uncertainty.array, hdr_uncertainty, name=hdu_uncertainty
            )
            hdus.append(hduUncert)

        if hdu_flags and self.flags:
            raise NotImplementedError(
                "adding the flags to a HDU is not supported at this time."
            )

        if hdu_psf and self.psf is not None:
            # The PSF is an image, so write it as a separate ImageHDU.
            hdu_psf = fits.ImageHDU(self.psf, name=hdu_psf)
            hdus.append(hdu_psf)

        return fits.HDUList(hdus)

    def copy(self):
        """
        Return a copy of the CCDData object.
        """
        return self.__class__(self, copy=True)

    add = _arithmetic(np.add)(NDDataArray.add)
    subtract = _arithmetic(np.subtract)(NDDataArray.subtract)
    multiply = _arithmetic(np.multiply)(NDDataArray.multiply)
    divide = _arithmetic(np.true_divide)(NDDataArray.divide)

    def _insert_in_metadata_fits_safe(self, key, value):
        """
        Insert key/value pair into metadata in a way that FITS can serialize.

        Parameters
        ----------
        key : str
            Key to be inserted in dictionary.

        value : str or None
            Value to be inserted.

        Notes
        -----
        This addresses a shortcoming of the FITS standard. There are length
        restrictions on both the ``key`` (8 characters) and ``value`` (72
        characters) in the FITS standard. There is a convention for handling
        long keywords and a convention for handling long values, but the
        two conventions cannot be used at the same time.

        This addresses that case by checking the length of the ``key`` and
        ``value`` and, if necessary, shortening the key.
        """
        if len(key) > 8 and len(value) > 72:
            short_name = key[:8]
            self.meta[f"HIERARCH {key.upper()}"] = (
                short_name,
                f"Shortened name for {key}",
            )
            self.meta[short_name] = value
        else:
            self.meta[key] = value

    # A dictionary mapping "known" invalid fits unit
    known_invalid_fits_unit_strings = {
        "ELECTRONS/S": u.electron / u.s,
        "ELECTRONS": u.electron,
        "electrons": u.electron,
    }


# These need to be importable by the tests...
_KEEP_THESE_KEYWORDS_IN_HEADER = ["JD-OBS", "MJD-OBS", "DATE-OBS"]
_PCs = {"PC1_1", "PC1_2", "PC2_1", "PC2_2"}
_CDs = {"CD1_1", "CD1_2", "CD2_1", "CD2_2"}


def _generate_wcs_and_update_header(hdr):
    """
    Generate a WCS object from a header and remove the WCS-specific
    keywords from the header.

    Parameters
    ----------
    hdr : astropy.io.fits.header or other dict-like

    Returns
    -------
    new_header, wcs
    """
    # Try constructing a WCS object.
    try:
        wcs = WCS(hdr)
    except Exception as exc:
        # Normally WCS only raises Warnings and doesn't fail but in rare
        # cases (malformed header) it could fail...
        log.info(
            "An exception happened while extracting WCS information from "
            f"the Header.\n{type(exc).__name__}: {str(exc)}"
        )
        return hdr, None
    # Test for success by checking to see if the wcs ctype has a non-empty
    # value, return None for wcs if ctype is empty.
    if not wcs.wcs.ctype[0]:
        return (hdr, None)

    new_hdr = hdr.copy()
    # If the keywords below are in the header they are also added to WCS.
    # It seems like they should *not* be removed from the header, though.

    wcs_header = wcs.to_header(relax=True)
    for k in wcs_header:
        if k not in _KEEP_THESE_KEYWORDS_IN_HEADER:
            new_hdr.remove(k, ignore_missing=True)

    # Check that this does not result in an inconsistent header WCS if the WCS
    # is converted back to a header.

    if (_PCs & set(wcs_header)) and (_CDs & set(new_hdr)):
        # The PCi_j representation is used by the astropy.wcs object,
        # so CDi_j keywords were not removed from new_hdr. Remove them now.
        for cd in _CDs:
            new_hdr.remove(cd, ignore_missing=True)

    # The other case -- CD in the header produced by astropy.wcs -- should
    # never happen based on [1], which computes the matrix in PC form.
    # [1]: https://github.com/astropy/astropy/blob/1cf277926d3598dd672dd528504767c37531e8c9/cextern/wcslib/C/wcshdr.c#L596
    #
    # The test test_ccddata.test_wcs_keyword_removal_for_wcs_test_files() does
    # check for the possibility that both PC and CD are present in the result
    # so if the implementation of to_header changes in wcslib in the future
    # then the tests should catch it, and then this code will need to be
    # updated.

    # We need to check for any SIP coefficients that got left behind if the
    # header has SIP.
    if wcs.sip is not None:
        keyword = "{}_{}_{}"
        polynomials = ["A", "B", "AP", "BP"]
        for poly in polynomials:
            order = wcs.sip.__getattribute__(f"{poly.lower()}_order")
            for i, j in itertools.product(range(order), repeat=2):
                new_hdr.remove(keyword.format(poly, i, j), ignore_missing=True)

    return (new_hdr, wcs)


def fits_ccddata_reader(
    filename,
    hdu=0,
    unit=None,
    hdu_uncertainty="UNCERT",
    hdu_mask="MASK",
    hdu_flags=None,
    key_uncertainty_type="UTYPE",
    hdu_psf="PSFIMAGE",
    **kwd,
):
    """
    Generate a CCDData object from a FITS file.

    Parameters
    ----------
    filename : str
        Name of fits file.

    hdu : int, str, tuple of (str, int), optional
        Index or other identifier of the Header Data Unit of the FITS
        file from which CCDData should be initialized. If zero and
        no data in the primary HDU, it will search for the first
        extension HDU with data. The header will be added to the primary HDU.
        Default is ``0``.

    unit : `~astropy.units.Unit`, optional
        Units of the image data. If this argument is provided and there is a
        unit for the image in the FITS header (the keyword ``BUNIT`` is used
        as the unit, if present), this argument is used for the unit.
        Default is ``None``.

    hdu_uncertainty : str or None, optional
        FITS extension from which the uncertainty should be initialized. If the
        extension does not exist the uncertainty of the CCDData is ``None``.
        Default is ``'UNCERT'``.

    hdu_mask : str or None, optional
        FITS extension from which the mask should be initialized. If the
        extension does not exist the mask of the CCDData is ``None``.
        Default is ``'MASK'``.

    hdu_flags : str or None, optional
        Currently not implemented.
        Default is ``None``.

    key_uncertainty_type : str, optional
        The header key name where the class name of the uncertainty  is stored
        in the hdu of the uncertainty (if any).
        Default is ``UTYPE``.

        .. versionadded:: 3.1

    hdu_psf : str or None, optional
        FITS extension from which the psf image should be initialized. If the
        extension does not exist the psf of the CCDData is `None`.

    kwd :
        Any additional keyword parameters are passed through to the FITS reader
        in :mod:`astropy.io.fits`; see Notes for additional discussion.

    Notes
    -----
    FITS files that contained scaled data (e.g. unsigned integer images) will
    be scaled and the keywords used to manage scaled data in
    :mod:`astropy.io.fits` are disabled.
    """
    unsupport_open_keywords = {
        "do_not_scale_image_data": "Image data must be scaled.",
        "scale_back": "Scale information is not preserved.",
    }
    for key, msg in unsupport_open_keywords.items():
        if key in kwd:
            prefix = f"unsupported keyword: {key}."
            raise TypeError(f"{prefix} {msg}")
    with fits.open(filename, **kwd) as hdus:
        hdr = hdus[hdu].header

        if hdu_uncertainty is not None and hdu_uncertainty in hdus:
            unc_hdu = hdus[hdu_uncertainty]
            stored_unc_name = unc_hdu.header.get(key_uncertainty_type, "None")
            # For compatibility reasons the default is standard deviation
            # uncertainty because files could have been created before the
            # uncertainty type was stored in the header.
            unc_type = _unc_name_to_cls.get(stored_unc_name, StdDevUncertainty)
            uncertainty = unc_type(unc_hdu.data)
        else:
            uncertainty = None

        if hdu_mask is not None and hdu_mask in hdus:
            # Mask is saved as uint but we want it to be boolean.
            mask = hdus[hdu_mask].data.astype(np.bool_)
        else:
            mask = None

        if hdu_flags is not None and hdu_flags in hdus:
            raise NotImplementedError("loading flags is currently not supported.")

        if hdu_psf is not None and hdu_psf in hdus:
            psf = hdus[hdu_psf].data
        else:
            psf = None

        # search for the first instance with data if
        # the primary header is empty.
        if hdu == 0 and hdus[hdu].data is None:
            for i in range(len(hdus)):
                if (
                    hdus.info(hdu)[i][3] == "ImageHDU"
                    and hdus.fileinfo(i)["datSpan"] > 0
                ):
                    hdu = i
                    comb_hdr = hdus[hdu].header.copy()
                    # Add header values from the primary header that aren't
                    # present in the extension header.
                    comb_hdr.extend(hdr, unique=True)
                    hdr = comb_hdr
                    log.info(f"first HDU with data is extension {hdu}.")
                    break

        if "bunit" in hdr:
            fits_unit_string = hdr["bunit"]
            # patch to handle FITS files using ADU for the unit instead of the
            # standard version of 'adu'
            if fits_unit_string.strip().lower() == "adu":
                fits_unit_string = fits_unit_string.lower()
        else:
            fits_unit_string = None

        if fits_unit_string:
            if unit is None:
                # Convert the BUNIT header keyword to a unit and if that's not
                # possible raise a meaningful error message.
                try:
                    kifus = CCDData.known_invalid_fits_unit_strings
                    if fits_unit_string in kifus:
                        fits_unit_string = kifus[fits_unit_string]
                    fits_unit_string = u.Unit(fits_unit_string)
                except ValueError:
                    raise ValueError(
                        f"The Header value for the key BUNIT ({fits_unit_string}) "
                        "cannot be interpreted as valid unit. To successfully read the "
                        "file as CCDData you can pass in a valid `unit` "
                        "argument explicitly or change the header of the FITS "
                        "file before reading it."
                    )
            else:
                log.info(
                    f"using the unit {unit} passed to the FITS reader instead "
                    f"of the unit {fits_unit_string} in the FITS file."
                )

        use_unit = unit or fits_unit_string
        hdr, wcs = _generate_wcs_and_update_header(hdr)

        return CCDData(
            hdus[hdu].data,
            meta=hdr,
            unit=use_unit,
            mask=mask,
            uncertainty=uncertainty,
            wcs=wcs,
            psf=psf,
        )


def fits_ccddata_writer(
    ccd_data,
    filename,
    hdu_mask="MASK",
    hdu_uncertainty="UNCERT",
    hdu_flags=None,
    key_uncertainty_type="UTYPE",
    as_image_hdu=False,
    hdu_psf="PSFIMAGE",
    **kwd,
):
    """
    Write CCDData object to FITS file.

    Parameters
    ----------
    ccd_data : CCDData
        Object to write.

    filename : str
        Name of file.

    hdu_mask, hdu_uncertainty, hdu_flags, hdu_psf : str or None, optional
        If it is a string append this attribute to the HDUList as
        `~astropy.io.fits.ImageHDU` with the string as extension name.
        Flags are not supported at this time. If ``None`` this attribute
        is not appended.
        Default is ``'MASK'`` for mask, ``'UNCERT'`` for uncertainty,
        ``'PSFIMAGE'`` for psf, and `None` for flags.

    key_uncertainty_type : str, optional
        The header key name for the class name of the uncertainty (if any)
        that is used to store the uncertainty type in the uncertainty hdu.
        Default is ``UTYPE``.

        .. versionadded:: 3.1

    as_image_hdu : bool
        If this option is `True`, the first item of the returned
        `~astropy.io.fits.HDUList` is a `~astropy.io.fits.ImageHDU`, instead of
        the default `~astropy.io.fits.PrimaryHDU`.

    kwd :
        All additional keywords are passed to :py:mod:`astropy.io.fits`

    Raises
    ------
    ValueError
        - If ``self.mask`` is set but not a `numpy.ndarray`.
        - If ``self.uncertainty`` is set but not a
          `~astropy.nddata.StdDevUncertainty`.
        - If ``self.uncertainty`` is set but has another unit then
          ``self.data``.

    NotImplementedError
        Saving flags is not supported.
    """
    hdu = ccd_data.to_hdu(
        hdu_mask=hdu_mask,
        hdu_uncertainty=hdu_uncertainty,
        key_uncertainty_type=key_uncertainty_type,
        hdu_flags=hdu_flags,
        as_image_hdu=as_image_hdu,
        hdu_psf=hdu_psf,
    )
    if as_image_hdu:
        hdu.insert(0, fits.PrimaryHDU())
    hdu.writeto(filename, **kwd)


with registry.delay_doc_updates(CCDData):
    registry.register_reader("fits", CCDData, fits_ccddata_reader)
    registry.register_writer("fits", CCDData, fits_ccddata_writer)
    registry.register_identifier("fits", CCDData, fits.connect.is_fits)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Regression tests for coordinates-related bugs that don't have an obvious other
place to live
"""

import copy
import io
from contextlib import nullcontext

import numpy as np
import pytest
from erfa import ErfaWarning

from astropy import units as u
from astropy.coordinates import (
    CIRS,
    FK4,
    GCRS,
    HCRS,
    ICRS,
    ITRS,
    AltAz,
    BaseCoordinateFrame,
    CartesianDifferential,
    CartesianRepresentation,
    CylindricalDifferential,
    CylindricalRepresentation,
    EarthLocation,
    FK4NoETerms,
    FunctionTransform,
    GeocentricMeanEcliptic,
    Latitude,
    Longitude,
    QuantityAttribute,
    SkyCoord,
    SphericalRepresentation,
    UnitSphericalRepresentation,
    get_body,
    get_sun,
)
from astropy.coordinates.sites import _GREENWICH
from astropy.table import Table
from astropy.tests.helper import assert_quantity_allclose
from astropy.time import Time
from astropy.units import allclose as quantity_allclose
from astropy.utils import iers
from astropy.utils.compat.optional_deps import HAS_SCIPY


def test_regression_5085():
    """
    PR #5085 was put in place to fix the following issue.

    Issue: https://github.com/astropy/astropy/issues/5069
    At root was the transformation of Ecliptic coordinates with
    non-scalar times.
    """
    # Note: for regression test, we need to be sure that we use UTC for the
    # epoch, even though more properly that should be TT; but the "expected"
    # values were calculated using that.
    j2000 = Time("J2000", scale="utc")
    times = Time(["2015-08-28 03:30", "2015-09-05 10:30", "2015-09-15 18:35"])
    latitudes = Latitude([3.9807075, -5.00733806, 1.69539491] * u.deg)
    longitudes = Longitude([311.79678613, 72.86626741, 199.58698226] * u.deg)
    distances = u.Quantity([0.00243266, 0.0025424, 0.00271296] * u.au)
    coo = GeocentricMeanEcliptic(
        lat=latitudes, lon=longitudes, distance=distances, obstime=times, equinox=times
    )
    # expected result
    ras = Longitude([310.50095400, 314.67109920, 319.56507428] * u.deg)
    decs = Latitude([-18.25190443, -17.1556676, -15.71616522] * u.deg)
    distances = u.Quantity([1.78309901, 1.710874, 1.61326649] * u.au)
    expected_result = GCRS(
        ra=ras, dec=decs, distance=distances, obstime=j2000
    ).cartesian.xyz
    actual_result = coo.transform_to(GCRS(obstime=j2000)).cartesian.xyz
    assert_quantity_allclose(expected_result, actual_result)


def test_regression_3920():
    """
    Issue: https://github.com/astropy/astropy/issues/3920
    """
    loc = EarthLocation.from_geodetic(0 * u.deg, 0 * u.deg, 0)
    time = Time("2010-1-1")

    aa = AltAz(location=loc, obstime=time)
    sc = SkyCoord(10 * u.deg, 3 * u.deg)
    assert sc.transform_to(aa).shape == ()
    # That part makes sense: the input is a scalar so the output is too

    sc2 = SkyCoord(10 * u.deg, 3 * u.deg, 1 * u.AU)
    assert sc2.transform_to(aa).shape == ()
    # in 3920 that assert fails, because the shape is (1,)

    # check that the same behavior occurs even if transform is from low-level classes
    icoo = ICRS(sc.data)
    icoo2 = ICRS(sc2.data)
    assert icoo.transform_to(aa).shape == ()
    assert icoo2.transform_to(aa).shape == ()


def test_regression_3938():
    """
    Issue: https://github.com/astropy/astropy/issues/3938
    """
    # Set up list of targets - we don't use `from_name` here to avoid
    # remote_data requirements, but it does the same thing
    # vega = SkyCoord.from_name('Vega')
    vega = SkyCoord(279.23473479 * u.deg, 38.78368896 * u.deg)
    # capella = SkyCoord.from_name('Capella')
    capella = SkyCoord(79.17232794 * u.deg, 45.99799147 * u.deg)
    # sirius = SkyCoord.from_name('Sirius')
    sirius = SkyCoord(101.28715533 * u.deg, -16.71611586 * u.deg)
    targets = [vega, capella, sirius]

    # Feed list of targets into SkyCoord
    combined_coords = SkyCoord(targets)

    # Set up AltAz frame
    time = Time("2012-01-01 00:00:00")
    location = EarthLocation("10d", "45d", 0)
    aa = AltAz(location=location, obstime=time)

    combined_coords.transform_to(aa)
    # in 3938 the above yields ``UnitConversionError: '' (dimensionless) and 'pc' (length) are not convertible``


def test_regression_3998():
    """
    Issue: https://github.com/astropy/astropy/issues/3998
    """
    time = Time("2012-01-01 00:00:00")
    assert time.isscalar

    sun = get_sun(time)
    assert sun.isscalar
    # in 3998, the above yields False - `sun` is a length-1 vector

    assert sun.obstime is time


def test_regression_4033():
    """
    Issue: https://github.com/astropy/astropy/issues/4033
    """
    # alb = SkyCoord.from_name('Albireo')
    alb = SkyCoord(292.68033548 * u.deg, 27.95968007 * u.deg)
    alb_wdist = SkyCoord(alb, distance=133 * u.pc)

    # de = SkyCoord.from_name('Deneb')
    de = SkyCoord(310.35797975 * u.deg, 45.28033881 * u.deg)
    de_wdist = SkyCoord(de, distance=802 * u.pc)

    aa = AltAz(
        location=EarthLocation(lat=45 * u.deg, lon=0 * u.deg), obstime="2010-1-1"
    )
    deaa = de.transform_to(aa)
    albaa = alb.transform_to(aa)
    alb_wdistaa = alb_wdist.transform_to(aa)
    de_wdistaa = de_wdist.transform_to(aa)

    # these work fine
    sepnod = deaa.separation(albaa)
    sepwd = deaa.separation(alb_wdistaa)
    assert_quantity_allclose(sepnod, 22.2862 * u.deg, rtol=1e-6)
    assert_quantity_allclose(sepwd, 22.2862 * u.deg, rtol=1e-6)
    # parallax should be present when distance added
    assert np.abs(sepnod - sepwd) > 1 * u.marcsec

    # in 4033, the following fail with a recursion error
    assert_quantity_allclose(
        de_wdistaa.separation(alb_wdistaa), 22.2862 * u.deg, rtol=1e-3
    )
    assert_quantity_allclose(alb_wdistaa.separation(deaa), 22.2862 * u.deg, rtol=1e-3)


@pytest.mark.skipif(not HAS_SCIPY, reason="No Scipy")
def test_regression_4082():
    """
    Issue: https://github.com/astropy/astropy/issues/4082
    """
    from astropy.coordinates import search_around_3d, search_around_sky

    cat = SkyCoord([10.076, 10.00455], [18.54746, 18.54896], unit="deg")
    search_around_sky(cat[0:1], cat, seplimit=u.arcsec * 60, storekdtree=False)
    # in the issue, this raises a TypeError

    # also check 3d for good measure, although it's not really affected by this bug directly
    cat3d = SkyCoord(
        [10.076, 10.00455] * u.deg,
        [18.54746, 18.54896] * u.deg,
        distance=[0.1, 1.5] * u.kpc,
    )
    search_around_3d(cat3d[0:1], cat3d, 1 * u.kpc, storekdtree=False)


def test_regression_4210():
    """
    Issue: https://github.com/astropy/astropy/issues/4210
    Related PR with actual change: https://github.com/astropy/astropy/pull/4211
    """
    crd = SkyCoord(0 * u.deg, 0 * u.deg, distance=1 * u.AU)
    ecl = crd.geocentricmeanecliptic
    # bug was that "lambda", which at the time was the name of the geocentric
    # ecliptic longitude, is a reserved keyword. So this just makes sure the
    # new name is are all valid
    ecl.lon

    # and for good measure, check the other ecliptic systems are all the same
    # names for their attributes
    from astropy.coordinates.builtin_frames import ecliptic

    for frame_name in ecliptic.__all__:
        eclcls = getattr(ecliptic, frame_name)
        eclobj = eclcls(1 * u.deg, 2 * u.deg, 3 * u.AU)

        eclobj.lat
        eclobj.lon
        eclobj.distance


def test_regression_futuretimes_4302():
    """
    Checks that an error is not raised for future times not covered by IERS
    tables (at least in a simple transform like CIRS->ITRS that simply requires
    the UTC<->UT1 conversion).

    Relevant comment: https://github.com/astropy/astropy/pull/4302#discussion_r44836531
    """
    # this is an ugly hack to get the warning to show up even if it has already
    # appeared
    from astropy.coordinates.builtin_frames import utils
    from astropy.utils.exceptions import AstropyWarning

    if hasattr(utils, "__warningregistry__"):
        utils.__warningregistry__.clear()

    # check that out-of-range warning appears among any other warnings.  If
    # tests are run with --remote-data then the IERS table will be an instance
    # of IERS_Auto which is assured of being "fresh".  In this case getting
    # times outside the range of the table does not raise an exception.  Only
    # if using IERS_B (which happens without --remote-data, i.e. for all CI
    # testing) do we expect another warning.
    if isinstance(iers.earth_orientation_table.get(), iers.IERS_B):
        ctx1 = pytest.warns(
            AstropyWarning,
            match=r"\(some\) times are outside of range covered by IERS table.*",
        )
    else:
        ctx1 = nullcontext()

    with (
        ctx1,
        pytest.warns(ErfaWarning, match=".*dubious year.*"),
        pytest.warns(AstropyWarning, match=".*times after IERS data is valid.*"),
    ):
        future_time = Time("2511-5-1")
        c = CIRS(1 * u.deg, 2 * u.deg, obstime=future_time)
        with iers.conf.set_temp("auto_max_age", None):
            c.transform_to(ITRS(obstime=future_time))


def test_regression_4996():
    # this part is the actual regression test
    deltat = np.linspace(-12, 12, 1000) * u.hour
    times = Time("2012-7-13 00:00:00") + deltat
    suncoo = get_sun(times)
    assert suncoo.shape == (len(times),)

    # and this is an additional test to make sure more complex arrays work
    times2 = Time("2012-7-13 00:00:00") + deltat.reshape(10, 20, 5)
    suncoo2 = get_sun(times2)
    assert suncoo2.shape == times2.shape

    # this is intentionally not allclose - they should be *exactly* the same
    assert np.all(suncoo.ra.ravel() == suncoo2.ra.ravel())


def test_regression_4293():
    """Really just an extra test on FK4 no e, after finding that the units
    were not always taken correctly.  This test is against explicitly doing
    the transformations on pp170 of Explanatory Supplement to the Astronomical
    Almanac (Seidelmann, 2005).

    See https://github.com/astropy/astropy/pull/4293#issuecomment-234973086
    """
    # Check all over sky, but avoiding poles (note that FK4 did not ignore
    # e terms within 10∘ of the poles...  see p170 of explan.supp.).
    ra, dec = np.meshgrid(np.arange(0, 359, 45), np.arange(-80, 81, 40))
    fk4 = FK4(ra.ravel() * u.deg, dec.ravel() * u.deg)

    Dc = -0.065838 * u.arcsec
    Dd = +0.335299 * u.arcsec
    # Dc * tan(obliquity), as given on p.170
    Dctano = -0.028553 * u.arcsec

    fk4noe_dec = (
        fk4.dec
        - (Dd * np.cos(fk4.ra) - Dc * np.sin(fk4.ra)) * np.sin(fk4.dec)
        - Dctano * np.cos(fk4.dec)
    )
    fk4noe_ra = fk4.ra - (Dc * np.cos(fk4.ra) + Dd * np.sin(fk4.ra)) / np.cos(fk4.dec)

    fk4noe = fk4.transform_to(FK4NoETerms())
    # Tolerance here just set to how well the coordinates match, which is much
    # better than the claimed accuracy of <1 mas for this first-order in
    # v_earth/c approximation.
    # Interestingly, if one divides by np.cos(fk4noe_dec) in the ra correction,
    # the match becomes good to 2 μas.
    assert_quantity_allclose(fk4noe.ra, fk4noe_ra, atol=11.0 * u.uas, rtol=0)
    assert_quantity_allclose(fk4noe.dec, fk4noe_dec, atol=3.0 * u.uas, rtol=0)


def test_regression_4926():
    times = Time("2010-01-1") + np.arange(20) * u.day

    # this is the regression test
    moon = get_body("moon", times, _GREENWICH)

    # this is an additional test to make sure the GCRS->ICRS transform works for complex shapes
    moon.transform_to(ICRS())

    # and some others to increase coverage of transforms
    moon.transform_to(HCRS(obstime="J2000"))
    moon.transform_to(HCRS(obstime=times))


def test_regression_5209():
    "check that distances are not lost on SkyCoord init"
    time = Time("2015-01-01")
    moon = get_body("moon", time)
    new_coord = SkyCoord([moon])
    assert_quantity_allclose(new_coord[0].distance, moon.distance)


def test_regression_5133():
    N = 1000
    np.random.seed(12345)
    lon = np.random.uniform(-10, 10, N) * u.deg
    lat = np.random.uniform(50, 52, N) * u.deg
    alt = np.random.uniform(0, 10.0, N) * u.km

    time = Time("2010-1-1")

    objects = EarthLocation.from_geodetic(lon, lat, height=alt)
    itrs_coo = objects.get_itrs(time)

    homes = [
        EarthLocation.from_geodetic(lon=-1 * u.deg, lat=52 * u.deg, height=h)
        for h in (0, 1000, 10000) * u.km
    ]

    altaz_frames = [AltAz(obstime=time, location=h) for h in homes]
    altaz_coos = [itrs_coo.transform_to(f) for f in altaz_frames]

    # they should all be different
    for coo in altaz_coos[1:]:
        assert not quantity_allclose(coo.az, coo.az[0])
        assert not quantity_allclose(coo.alt, coo.alt[0])


def test_itrs_vals_5133():
    """
    Test to check if alt-az calculations respect height of observer

    Because ITRS is geocentric and includes aberration, an object that
    appears 'straight up' to a geocentric observer (ITRS) won't be
    straight up to a topocentric observer - see

    https://github.com/astropy/astropy/issues/10983

    This is worse for small height above the Earth, which is why this test
    uses large distances.
    """
    time = Time("2010-1-1")
    height = 500000.0 * u.km
    el = EarthLocation.from_geodetic(lon=20 * u.deg, lat=45 * u.deg, height=height)

    lons = [20, 30, 20] * u.deg
    lats = [44, 45, 45] * u.deg
    alts = u.Quantity([height, height, 10 * height])
    coos = [
        EarthLocation.from_geodetic(lon, lat, height=alt).get_itrs(time)
        for lon, lat, alt in zip(lons, lats, alts)
    ]

    aaf = AltAz(obstime=time, location=el)
    aacs = [coo.transform_to(aaf) for coo in coos]

    assert all(coo.isscalar for coo in aacs)

    # the ~1 degree tolerance is b/c aberration makes it not exact
    assert_quantity_allclose(aacs[0].az, 180 * u.deg, atol=1 * u.deg)
    assert aacs[0].alt < 0 * u.deg
    assert aacs[0].distance > 5000 * u.km

    # it should *not* actually be 90 degrees, b/c constant latitude is not
    # straight east anywhere except the equator... but should be close-ish
    assert_quantity_allclose(aacs[1].az, 90 * u.deg, atol=5 * u.deg)
    assert aacs[1].alt < 0 * u.deg
    assert aacs[1].distance > 5000 * u.km

    assert_quantity_allclose(aacs[2].alt, 90 * u.deg, atol=1 * u.arcminute)
    assert_quantity_allclose(aacs[2].distance, 9 * height)


def test_regression_simple_5133():
    """
    Simple test to check if alt-az calculations respect height of observer

    Because ITRS is geocentric and includes aberration, an object that
    appears 'straight up' to a geocentric observer (ITRS) won't be
    straight up to a topocentric observer - see

    https://github.com/astropy/astropy/issues/10983

    This is why we construct a topocentric GCRS SkyCoord before calculating AltAz
    """
    t = Time("J2010")
    obj = EarthLocation(-1 * u.deg, 52 * u.deg, height=[10.0, 0.0] * u.km)
    home = EarthLocation(-1 * u.deg, 52 * u.deg, height=5.0 * u.km)

    obsloc_gcrs, obsvel_gcrs = home.get_gcrs_posvel(t)
    gcrs_geo = obj.get_itrs(t).transform_to(GCRS(obstime=t))
    obsrepr = home.get_itrs(t).transform_to(GCRS(obstime=t)).cartesian
    topo_gcrs_repr = gcrs_geo.cartesian - obsrepr
    topocentric_gcrs_frame = GCRS(
        obstime=t, obsgeoloc=obsloc_gcrs, obsgeovel=obsvel_gcrs
    )
    gcrs_topo = topocentric_gcrs_frame.realize_frame(topo_gcrs_repr)
    aa = gcrs_topo.transform_to(AltAz(obstime=t, location=home))

    # az is more-or-less undefined for straight up or down
    assert_quantity_allclose(aa.alt, [90, -90] * u.deg, rtol=1e-7)
    assert_quantity_allclose(aa.distance, 5 * u.km)


def test_regression_5743():
    sc = SkyCoord(
        [5, 10], [20, 30], unit=u.deg, obstime=["2017-01-01T00:00", "2017-01-01T00:10"]
    )
    assert sc[0].obstime.shape == ()


def test_regression_5889_5890():
    # ensure we can represent all Representations and transform to ND frames
    times = Time("2017-03-20T12:00:00") + np.linspace(-2, 2, 3) * u.hour
    moon = get_body("moon", times, location=_GREENWICH)
    targets = SkyCoord([350.7 * u.deg, 260.7 * u.deg], [18.4 * u.deg, 22.4 * u.deg])
    targs2d = targets[:, np.newaxis]
    targs2d.transform_to(moon)


def test_regression_6236():
    # sunpy changes its representation upon initialisation of a frame,
    # including via `realize_frame`. Ensure this works.
    class MyFrame(BaseCoordinateFrame):
        default_representation = CartesianRepresentation
        my_attr = QuantityAttribute(default=0, unit=u.m)

    class MySpecialFrame(MyFrame):
        def __init__(self, *args, **kwargs):
            _rep_kwarg = kwargs.get("representation_type")
            super().__init__(*args, **kwargs)
            if not _rep_kwarg:
                self.representation_type = self.default_representation
                self._data = self.data.represent_as(self.representation_type)

    rep1 = UnitSphericalRepresentation([0.0, 1] * u.deg, [2.0, 3.0] * u.deg)
    rep2 = SphericalRepresentation(
        [10.0, 11] * u.deg, [12.0, 13.0] * u.deg, [14.0, 15.0] * u.kpc
    )
    mf1 = MyFrame(rep1, my_attr=1.0 * u.km)
    mf2 = mf1.realize_frame(rep2)
    # Normally, data is stored as is, but the representation gets set to a
    # default, even if a different representation instance was passed in.
    # realize_frame should do the same. Just in case, check attrs are passed.
    assert mf1.data is rep1
    assert mf2.data is rep2
    assert mf1.representation_type is CartesianRepresentation
    assert mf2.representation_type is CartesianRepresentation
    assert mf2.my_attr == mf1.my_attr
    # It should be independent of whether I set the representation explicitly
    mf3 = MyFrame(rep1, my_attr=1.0 * u.km, representation_type="unitspherical")
    mf4 = mf3.realize_frame(rep2)
    assert mf3.data is rep1
    assert mf4.data is rep2
    assert mf3.representation_type is UnitSphericalRepresentation
    assert mf4.representation_type is CartesianRepresentation
    assert mf4.my_attr == mf3.my_attr
    # This should be enough to help sunpy, but just to be sure, a test
    # even closer to what is done there, i.e., transform the representation.
    msf1 = MySpecialFrame(rep1, my_attr=1.0 * u.km)
    msf2 = msf1.realize_frame(rep2)
    assert msf1.data is not rep1  # Gets transformed to Cartesian.
    assert msf2.data is not rep2
    assert type(msf1.data) is CartesianRepresentation
    assert type(msf2.data) is CartesianRepresentation
    assert msf1.representation_type is CartesianRepresentation
    assert msf2.representation_type is CartesianRepresentation
    assert msf2.my_attr == msf1.my_attr
    # And finally a test where the input is not transformed.
    msf3 = MySpecialFrame(rep1, my_attr=1.0 * u.km, representation_type="unitspherical")
    msf4 = msf3.realize_frame(rep2)
    assert msf3.data is rep1
    assert msf4.data is not rep2
    assert msf3.representation_type is UnitSphericalRepresentation
    assert msf4.representation_type is CartesianRepresentation
    assert msf4.my_attr == msf3.my_attr


def test_gcrs_itrs_cartesian_repr():
    # issue 6436: transformation failed if coordinate representation was
    # Cartesian
    gcrs = GCRS(
        CartesianRepresentation((859.07256, -4137.20368, 5295.56871), unit="km"),
        representation_type="cartesian",
    )
    gcrs.transform_to(ITRS())


def test_regression_6446():
    # this succeeds even before 6446:
    sc1 = SkyCoord([1, 2], [3, 4], unit="deg")
    t1 = Table([sc1])
    sio1 = io.StringIO()
    t1.write(sio1, format="ascii.ecsv")

    # but this fails due to the 6446 bug
    c1 = SkyCoord(1, 3, unit="deg")
    c2 = SkyCoord(2, 4, unit="deg")
    sc2 = SkyCoord([c1, c2])
    t2 = Table([sc2])
    sio2 = io.StringIO()
    t2.write(sio2, format="ascii.ecsv")

    assert sio1.getvalue() == sio2.getvalue()


def test_regression_6597():
    frame_name = "galactic"
    c1 = SkyCoord(1, 3, unit="deg", frame=frame_name)
    c2 = SkyCoord(2, 4, unit="deg", frame=frame_name)
    sc1 = SkyCoord([c1, c2])

    assert sc1.frame.name == frame_name


def test_regression_6597_2():
    """
    This tests the more subtle flaw that #6597 indirectly uncovered: that even
    in the case that the frames are ra/dec, they still might be the wrong *kind*
    """
    frame = FK4(equinox="J1949")
    c1 = SkyCoord(1, 3, unit="deg", frame=frame)
    c2 = SkyCoord(2, 4, unit="deg", frame=frame)
    sc1 = SkyCoord([c1, c2])

    assert sc1.frame.name == frame.name


def test_regression_6697():
    """
    Test for regression of a bug in get_gcrs_posvel that introduced errors at the 1m/s level.

    Comparison data is derived from calculation in PINT
    https://github.com/nanograv/PINT/blob/master/pint/erfautils.py
    """
    pint_vels = CartesianRepresentation(
        348.63632871, -212.31704928, -0.60154936, unit=u.m / u.s
    )
    location = EarthLocation(
        5327448.9957829, -1718665.73869569, 3051566.90295403, unit=u.m
    )
    t = Time(2458036.161966612, format="jd")
    obsgeopos, obsgeovel = location.get_gcrs_posvel(t)
    delta = (obsgeovel - pint_vels).norm()
    assert delta < 1 * u.cm / u.s


def test_regression_8138():
    sc = SkyCoord(1 * u.deg, 2 * u.deg)
    newframe = GCRS()
    sc2 = sc.transform_to(newframe)
    assert newframe.is_equivalent_frame(sc2.frame)


def test_regression_8276():
    from astropy.coordinates import baseframe

    class MyFrame(BaseCoordinateFrame):
        a = QuantityAttribute(unit=u.m)

    # we save the transform graph so that it doesn't accidentally mess with other tests
    old_transform_graph = baseframe.frame_transform_graph
    try:
        baseframe.frame_transform_graph = copy.copy(baseframe.frame_transform_graph)

        # as reported in 8276, this previously failed right here because
        # registering the transform tries to create a frame attribute
        @baseframe.frame_transform_graph.transform(FunctionTransform, MyFrame, AltAz)
        def trans(my_frame_coord, altaz_frame):
            pass

        # should also be able to *create* the Frame at this point
        MyFrame()
    finally:
        baseframe.frame_transform_graph = old_transform_graph


def test_regression_8615():
    # note this is a "higher-level" symptom of the problem that a test now moved
    # to pyerfa (erfa/tests/test_erfa:test_float32_input) is testing for, but we keep
    # it here as well due to being a more practical version of the issue.

    crf = CartesianRepresentation(np.array([3, 0, 4], dtype=float) * u.pc)
    srf = SphericalRepresentation.from_cartesian(crf)  # does not error in 8615

    cr = CartesianRepresentation(np.array([3, 0, 4], dtype="f4") * u.pc)
    sr = SphericalRepresentation.from_cartesian(cr)  # errors in 8615

    assert_quantity_allclose(sr.distance, 5 * u.pc)
    assert_quantity_allclose(srf.distance, 5 * u.pc)


def test_regression_8924():
    """This checks that the ValueError in
    BaseRepresentation._re_represent_differentials is raised properly
    """
    # A case where the representation has a 's' differential, but we try to
    # re-represent only with an 's2' differential
    rep = CartesianRepresentation(1, 2, 3, unit=u.kpc)
    dif = CartesianDifferential(4, 5, 6, u.km / u.s)
    rep = rep.with_differentials(dif)

    with pytest.raises(ValueError):
        rep._re_represent_differentials(
            CylindricalRepresentation, {"s2": CylindricalDifferential}
        )


def test_regression_10092():
    """
    Check that we still get a proper motion even for SkyCoords without distance
    """
    c = SkyCoord(
        l=10 * u.degree,
        b=45 * u.degree,
        pm_l_cosb=34 * u.mas / u.yr,
        pm_b=-117 * u.mas / u.yr,
        frame="galactic",
        obstime=Time("1988-12-18 05:11:23.5"),
    )

    with pytest.warns(ErfaWarning, match='ERFA function "pmsafe" yielded .*'):
        newc = c.apply_space_motion(dt=10 * u.year)
    assert_quantity_allclose(
        newc.pm_l_cosb, 33.99980714 * u.mas / u.yr, atol=1.0e-5 * u.mas / u.yr
    )


def test_regression_10226():
    # Dictionary representation of SkyCoord should contain differentials.
    sc = SkyCoord(
        [270, 280] * u.deg,
        [30, 35] * u.deg,
        [10, 11] * u.pc,
        radial_velocity=[20, -20] * u.km / u.s,
    )
    sc_as_dict = sc.info._represent_as_dict()
    assert "radial_velocity" in sc_as_dict
    # But only the components that have been specified.
    assert "pm_dec" not in sc_as_dict


@pytest.mark.parametrize(
    "mjd", (52000, [52000], [[52000]], [52001, 52002], [[52001], [52002]])
)
def test_regression_10422(mjd):
    """
    Check that we can get a GCRS for a scalar EarthLocation and a
    size=1 non-scalar Time.
    """
    # Avoid trying to download new IERS data.
    with iers.earth_orientation_table.set(iers.IERS_B.open(iers.IERS_B_FILE)):
        t = Time(mjd, format="mjd", scale="tai")
        loc = EarthLocation(88258.0 * u.m, -4924882.2 * u.m, 3943729.0 * u.m)
        p, v = loc.get_gcrs_posvel(obstime=t)
        assert p.shape == v.shape == t.shape


@pytest.mark.remote_data
def test_regression_10291():
    """
    According to https://eclipse.gsfc.nasa.gov/OH/transit12.html,
    the minimum separation between Venus and the Sun during the 2012
    transit is 554 arcseconds for an observer at the Geocenter.

    If light deflection from the Sun is incorrectly applied, this increases
    to 557 arcseconds.
    """
    t = Time("2012-06-06 01:29:36")
    sun = get_body("sun", t)
    venus = get_body("venus", t)
    assert_quantity_allclose(
        venus.separation(sun), 554.427 * u.arcsecond, atol=0.001 * u.arcsecond
    )


@pytest.mark.parametrize("coord_cls", [ICRS, SkyCoord])
@pytest.mark.parametrize(
    "differential_type, diff_kwargs",
    [
        (None, {}),
        (
            "unitsphericalcoslat",
            {
                "pm_ra_cosdec": [40, 50] * u.mas / u.yr,
                "pm_dec": [60, 70] * u.mas / u.yr,
            },
        ),
        ("radial", {"radial_velocity": [80, 90] * u.km / u.s}),
    ],
)
@pytest.mark.parametrize("extra_kwargs", [{}, {"representation_type": "unitspherical"}])
def test_regression_16998(coord_cls, differential_type, diff_kwargs, extra_kwargs):
    """Direct tests of the underlying problem causing gh-16998.

    Note that the issue itself was of columns missing in data written to a file.
    That is now tested directly by the "icrs" column in
    astropy/io/tests/mixin_columns.py.
    Here, we test the underlying problem, that .info._represent_as_dict()
    did not return a full set of columns in some cases.
    """
    if extra_kwargs and differential_type:
        extra_kwargs["differential_type"] = differential_type
    coord = coord_cls([0, 10] * u.deg, [20, 30] * u.deg, **diff_kwargs, **extra_kwargs)
    expected_entries = {"ra", "dec", "representation_type"}
    if differential_type:
        expected_entries |= {"differential_type"} | set(diff_kwargs)
    if coord_cls is SkyCoord:
        expected_entries.add("frame")
    assert set(coord.info._represent_as_dict()) == expected_entries


def test_regression_17008():
    """Test that one can transform a SkyCoord with empty data to other frames.

    The underlying bug that caused the problem reported in gh-17008
    is tested in utils/iers/tests/test_iers.py::test_empty_mjd
    """
    s = SkyCoord([] * u.deg, [] * u.deg, obstime=Time([], format="iso"))
    itrs = s.itrs  # This failed before.
    assert itrs.size == 0



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import subprocess
import sys
from textwrap import dedent

import pytest

import astropy.constants as const
from astropy import astronomical_constants, physical_constants


def test_version_match():
    pversion = physical_constants.get()
    refpversion = const.h.__class__.__name__.lower()
    assert pversion == refpversion
    aversion = astronomical_constants.get()
    refaversion = const.M_sun.__class__.__name__.lower()
    assert aversion == refaversion


def test_previously_imported():
    with pytest.raises(RuntimeError):
        physical_constants.set("codata2018")

    with pytest.raises(RuntimeError):
        astronomical_constants.set("iau2015")


@pytest.mark.parametrize("version", sorted(set(physical_constants._versions.values())))
def test_physical_constants_versions(version):
    """Spot check that setting the different physical constants actually works.

    Here, checks the Rydberg constant as one that seems to change in every
    version (at least up to codata2022), and uses the electron charge to test
    that units derived from constants change as well.
    """
    cmd = dedent(f"""
    import astropy
    astropy.physical_constants.set({version!r})
    import astropy.constants as const
    from astropy.constants import {version} as {version}, codata2018 as codata2018
    assert const.Ryd.value == {version}.Ryd.value
    assert const.Ryd.uncertainty == {version}.Ryd.uncertainty
    if {version} != codata2018:
        assert {version}.Ryd.uncertainty != codata2018.Ryd.uncertainty
    import astropy.units as u
    assert u.eV.to(u.J) == {version}.e.value
    """)
    cp = subprocess.check_call([sys.executable, "-c", cmd])
    assert cp == 0



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test that Distribution works with classes other than ndarray and Quantity."""

import numpy as np
import pytest
from numpy.testing import assert_array_equal

import astropy.units as u
from astropy.coordinates import (
    Angle,
    CartesianDifferential,
    CartesianRepresentation,
    EarthLocation,
    Latitude,
    Longitude,
    SkyCoord,
    SphericalDifferential,
    SphericalRepresentation,
)
from astropy.coordinates.representation import (
    DIFFERENTIAL_CLASSES,
    REPRESENTATION_CLASSES,
)
from astropy.coordinates.tests.test_representation import (
    components_allclose,
    representation_equal,
)
from astropy.uncertainty import Distribution


def assert_representation_equal(rep1, rep2):
    __tracebackhide__ = True
    assert np.all(representation_equal(rep1, rep2))


def assert_representation_allclose(rep1, rep2):
    __tracebackhide__ = True
    result = True
    if type(rep1) is not type(rep2):
        return False
    if getattr(rep1, "_differentials", False):
        if rep1._differentials.keys() != rep2._differentials.keys():
            return False
        for key, diff1 in rep1._differentials.items():
            result &= components_allclose(diff1, rep2._differentials[key])
    elif getattr(rep2, "_differentials", False):
        return False

    return np.all(result & components_allclose(rep1, rep2))


class TestAngles:
    @classmethod
    def setup_class(cls):
        cls.a = np.arange(27.0).reshape(3, 9)
        cls.d = Distribution(cls.a)
        cls.q = cls.a << u.deg
        cls.dq = Distribution(cls.q)

    @pytest.mark.parametrize("angle_cls", [Angle, Longitude, Latitude])
    def test_as_input_for_angle(self, angle_cls):
        da = angle_cls(self.dq)
        assert isinstance(da, angle_cls)
        assert isinstance(da, Distribution)
        assert_array_equal(da.distribution, angle_cls(self.q))

    @pytest.mark.parametrize("angle_cls", [Angle, Longitude, Latitude])
    def test_using_angle_as_input(self, angle_cls):
        a = angle_cls(self.q)
        da = Distribution(a)
        assert isinstance(da, angle_cls)
        assert isinstance(da, Distribution)

    # Parametrize the unit to check the various branches in Latitude._validate_angles
    @pytest.mark.parametrize("dtype", ["f8", "f4"])
    @pytest.mark.parametrize(
        "value", [90 * u.deg, np.pi / 2 * u.radian, 90 * 60 * u.arcmin]
    )
    def test_at_limit_for_latitude(self, value, dtype):
        q = u.Quantity(value, dtype=dtype).reshape(1)
        qd = Distribution(q)
        ld = Latitude(qd)
        assert_array_equal(ld.distribution, Latitude(q))

    # Parametrize the unit in case Longitude._wrap_at becomes unit-dependent.
    @pytest.mark.parametrize("dtype", ["f8", "f4"])
    @pytest.mark.parametrize(
        "value", [360 * u.deg, 2 * np.pi * u.radian, 360 * 60 * u.arcmin]
    )
    def test_at_wrap_angle_for_longitude(self, value, dtype):
        q = u.Quantity(value, dtype=dtype).reshape(1)
        qd = Distribution(q)
        ld = Longitude(qd)
        assert_array_equal(ld.distribution, Longitude(q))
        assert np.all(ld.distribution == 0)

    @pytest.mark.parametrize("angle_cls", [Longitude, Latitude])
    def test_operation_gives_correct_subclass(self, angle_cls):
        # Lon and Lat always fall back to Angle
        da = angle_cls(self.dq)
        da2 = da + da
        assert isinstance(da, Angle)
        assert isinstance(da, Distribution)

    @pytest.mark.parametrize("angle_cls", [Longitude, Latitude])
    def test_pdfstd_gives_correct_subclass(self, angle_cls):
        # Lon and Lat always fall back to Angle
        da = angle_cls(self.dq)
        std = da.pdf_std()
        assert isinstance(std, Angle)
        assert_array_equal(std, Angle(self.q.std(-1)))

    def test_earthlocation_geocentric_distribution(self):
        x = y = z = self.a << u.km

        eloc = EarthLocation.from_geocentric(x=x, y=y, z=z)

        xd = Distribution(x)
        yd = Distribution(y)
        zd = Distribution(z)
        deloc = EarthLocation.from_geocentric(x=xd, y=yd, z=zd)

        assert isinstance(deloc.x, Distribution)
        assert_array_equal(np.median(eloc.x, axis=1), deloc.x.pdf_median())

    def test_earthlocation_geodetic_distribution(self):
        h = self.a << u.km
        eloc = EarthLocation.from_geodetic(lon=self.q, lat=self.q, height=h)

        hq = Distribution(h)
        deloc = EarthLocation.from_geodetic(lon=self.dq, lat=self.dq, height=hq)

        assert isinstance(deloc.x, Distribution)
        assert_array_equal(np.median(eloc.x, axis=1), deloc.x.pdf_median())


class TestRepresentation:
    @classmethod
    def setup_class(cls):
        cls.lon = Distribution(np.linspace(0.0, 360 * u.deg, 10, endpoint=False))
        cls.lat = Angle([-45.0, 0.0, 45.0], u.deg)
        cls.r = 6000 * u.km  # Sort of OK for Geodetic representations.
        cls.sph = SphericalRepresentation(
            cls.lon.distribution, cls.lat[:, np.newaxis], cls.r
        )
        cls.dsph = SphericalRepresentation(cls.lon, cls.lat, cls.r)

    def get_distribution(self, rep):
        return rep._apply(lambda x: getattr(x, "distribution", x[..., np.newaxis]))

    def test_cartesian(self):
        dcart = self.dsph.to_cartesian()
        cart = self.sph.to_cartesian()
        assert isinstance(dcart.x, Distribution)
        assert_array_equal(dcart.x.distribution, cart.x)
        assert_array_equal(dcart.y.distribution, cart.y)
        assert_array_equal(dcart.z.distribution, cart.z)

    def test_cartesian_roundtrip(self):
        roundtrip = SphericalRepresentation.from_cartesian(self.dsph.to_cartesian())
        assert_representation_allclose(self.get_distribution(roundtrip), self.sph)

    @pytest.mark.parametrize("rep_cls", REPRESENTATION_CLASSES.values())
    def test_other_reps(self, rep_cls):
        drep = self.dsph.represent_as(rep_cls)
        rep = self.sph.represent_as(rep_cls)
        assert_representation_equal(self.get_distribution(drep), rep)


class TestRepresentationWithDifferential(TestRepresentation):
    @classmethod
    def setup_class(cls):
        super().setup_class()
        cls.d_lon = Distribution(np.ones(10) << u.deg / u.hour)
        cls.d_lat = np.ones(cls.lat.shape) << u.deg / u.hour
        cls.d_r = 1 * u.m / u.s
        cls.d_sph = SphericalDifferential(
            cls.d_lon.distribution, cls.d_lat[:, np.newaxis], cls.d_r
        )
        cls.dd_sph = SphericalDifferential(cls.d_lon, cls.d_lat, cls.d_r)
        cls.sph = cls.sph.with_differentials({"s": cls.d_sph})
        cls.dsph = cls.dsph.with_differentials({"s": cls.dd_sph})

    def test_cartesian(self):
        dcart = self.dsph.represent_as(CartesianRepresentation, CartesianDifferential)
        cart = self.sph.represent_as(CartesianRepresentation, CartesianDifferential)
        assert isinstance(dcart.x, Distribution)
        assert_array_equal(dcart.x.distribution, cart.x)
        assert_array_equal(dcart.y.distribution, cart.y)
        assert_array_equal(dcart.z.distribution, cart.z)
        assert "s" in dcart._differentials
        dd_cart = dcart._differentials["s"]
        d_cart = cart._differentials["s"]
        assert_array_equal(dd_cart.d_x.distribution, d_cart.d_x)
        assert_array_equal(dd_cart.d_y.distribution, d_cart.d_y)
        assert_array_equal(dd_cart.d_z.distribution, d_cart.d_z)

    def test_cartesian_roundtrip(self):
        roundtrip = self.dsph.represent_as(
            CartesianRepresentation, CartesianDifferential
        ).represent_as(SphericalRepresentation, SphericalDifferential)
        assert_representation_allclose(self.get_distribution(roundtrip), self.sph)

    @pytest.mark.parametrize("diff_cls", DIFFERENTIAL_CLASSES.values())
    def test_other_reps(self, diff_cls):
        repr_cls = diff_cls.base_representation
        drep = self.dsph.represent_as(repr_cls, diff_cls)
        rep = self.sph.represent_as(repr_cls, diff_cls)
        assert_representation_equal(self.get_distribution(drep), rep)


class TestSkyCoord:
    @classmethod
    def setup_class(cls):
        cls.ra = Distribution(np.linspace(0.0, 360 * u.deg, 10, endpoint=False))
        cls.dec = Angle([-45.0, 0.0, 45.0], u.deg)
        cls.dsc = SkyCoord(cls.ra, cls.dec)
        cls.sc = SkyCoord(cls.ra.distribution, cls.dec[:, np.newaxis])

    def test_init(self):
        assert isinstance(self.dsc.ra, Distribution)
        assert not isinstance(self.dsc.dec, Distribution)

    @pytest.mark.parametrize("frame", ["fk5", "galactic"])
    def test_convert(self, frame):
        dconv = self.dsc.transform_to(frame)
        conv = self.sc.transform_to(frame)
        # Coordinate names are different for FK5 and galactic, so
        # just get them from the representation.
        assert_array_equal(dconv.data.lon.distribution, conv.data.lon)
        assert_array_equal(dconv.data.lat.distribution, conv.data.lat)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Combine 3 images to produce properly-scaled RGB images with arbitrary scaling.

The three images must be aligned and have the same pixel scale and size.
"""

import numpy as np

from astropy.visualization.interval import ManualInterval
from astropy.visualization.stretch import LinearStretch

_OUTPUT_IMAGE_FORMATS = [float, np.float64, np.uint8]

__all__ = ["make_rgb"]


class RGBImageMapping:
    """
    Class to map red, blue, green images into either a normalized float or
    an 8-bit image, by performing optional clipping and applying
    a scaling function to each band independently.

    Parameters
    ----------
    interval : `~astropy.visualization.BaseInterval` subclass instance or array-like, optional
        The interval object to apply to the data (either a single instance or
        an array for R, G, B). Default is
        `~astropy.visualization.ManualInterval`.
    stretch : `~astropy.visualization.BaseStretch` subclass instance, optional
        The stretch object to apply to the data. Default is
        `~astropy.visualization.LinearStretch`.

    """

    def __init__(
        self, interval=ManualInterval(vmin=0, vmax=None), stretch=LinearStretch()
    ):
        try:
            len(interval)
        except TypeError:
            interval = 3 * [interval]
        if len(interval) != 3:
            raise ValueError("please provide 1 or 3 instances for interval.")

        self.intervals = interval
        self.stretch = stretch

    def make_rgb_image(self, image_r, image_g, image_b, output_dtype=np.uint8):
        """
        Convert 3 arrays, image_r, image_g, and image_b into a RGB image,
        either as an 8-bit per-channel or normalized image.

        The input images can be int or float, and in any range or bit-depth,
        but must have the same shape (NxM).

        Parameters
        ----------
        image_r : ndarray
            Image to map to red.
        image_g : ndarray
            Image to map to green.
        image_b : ndarray
            Image to map to blue.
        output_dtype : numpy scalar type, optional
            Image output format. Default is np.uint8.

        Returns
        -------
        RGBimage : ndarray
            RGB color image with the specified format as an NxMx3 numpy array.

        """
        if output_dtype not in _OUTPUT_IMAGE_FORMATS:
            raise ValueError(f"'output_dtype' must be one of {_OUTPUT_IMAGE_FORMATS}!")

        image_r = np.asarray(image_r)
        image_g = np.asarray(image_g)
        image_b = np.asarray(image_b)

        if (image_r.shape != image_g.shape) or (image_g.shape != image_b.shape):
            msg = "The image shapes must match. r: {}, g: {} b: {}"
            raise ValueError(msg.format(image_r.shape, image_g.shape, image_b.shape))

        image_rgb = self.apply_mappings(image_r, image_g, image_b)
        if np.issubdtype(output_dtype, float):
            conv_images = self._convert_images_to_float(image_rgb, output_dtype)
        elif np.issubdtype(output_dtype, np.unsignedinteger):
            conv_images = self._convert_images_to_uint(image_rgb, output_dtype)

        return np.dstack(conv_images)

    def apply_mappings(self, image_r, image_g, image_b):
        """
        Apply mapping stretch and intervals to convert images image_r, image_g,
        and image_b to a triplet of normalized images.

        Parameters
        ----------
        image_r : ndarray
            Intensity of image to be mapped to red
        image_g : ndarray
            Intensity of image to be mapped to green.
        image_b : ndarray
            Intensity of image to be mapped to blue.

        Returns
        -------
        image_rgb : ndarray
            Triplet of mapped images based on the specified (per-band)
            intervals and the stretch function

        """
        image_rgb = [image_r, image_g, image_b]
        for i, img in enumerate(image_rgb):
            # Using syntax from mpl_normalize.ImageNormalize,
            # but NOT using that class to avoid dependency on matplotlib.

            # Define vmin and vmax
            vmin, vmax = self.intervals[i].get_limits(img)

            # copy because of in-place operations after
            img = np.array(img, copy=True, dtype=float)

            # Normalize based on vmin and vmax
            np.subtract(img, vmin, out=img)
            np.true_divide(img, vmax - vmin, out=img)

            # Clip to the 0 to 1 range
            img = np.clip(img, 0.0, 1.0, out=img)

            # Stretch values
            img = self.stretch(img, out=img, clip=False)

            image_rgb[i] = img

        return np.asarray(image_rgb)

    def _convert_images_to_float(self, image_rgb, output_dtype):
        """
        Convert a triplet of normalized images to float.
        """
        return image_rgb.astype(output_dtype)

    def _convert_images_to_uint(self, image_rgb, output_dtype):
        """
        Convert a triplet of normalized images to unsigned integer images
        """
        pixmax = float(np.iinfo(output_dtype).max)

        image_rgb *= pixmax

        return image_rgb.astype(output_dtype)


def make_rgb(
    image_r,
    image_g,
    image_b,
    interval=ManualInterval(vmin=0, vmax=None),
    stretch=LinearStretch(),
    filename=None,
    output_dtype=np.uint8,
):
    """
    Base class to return a Red/Green/Blue color image from 3 images using
    a specified stretch and interval, for each band *independently*.

    The input images can be int or float, and in any range or bit-depth,
    but must have the same shape (NxM).

    For a more detailed look at the use of this method, see the document
    :ref:`astropy:astropy-visualization-rgb`.

    Parameters
    ----------
    image_r : ndarray
        Image to map to red.
    image_g : ndarray
        Image to map to green.
    image_b : ndarray
        Image to map to blue.
    interval : `~astropy.visualization.BaseInterval` subclass instance or array-like, optional
        The interval object to apply to the data (either a single instance or
        an array for R, G, B). Default is
        `~astropy.visualization.ManualInterval` with vmin=0.
    stretch : `~astropy.visualization.BaseStretch` subclass instance, optional
        The stretch object to apply to the data. Default is
        `~astropy.visualization.LinearStretch`.
    filename : str, optional
        Write the resulting RGB image to a file (file type determined
        from extension).
    output_dtype : numpy scalar type, optional
        Image output data type. Default is np.uint8.

    Returns
    -------
    rgb : ndarray
        RGB (either float or integer with 8-bits per channel) color image
        as an NxMx3 numpy array.

    Notes
    -----
    This procedure of clipping and then scaling is similar to the DS9
    image algorithm (see the DS9 reference guide:
    http://ds9.si.edu/doc/ref/how.html).

    """
    map_ = RGBImageMapping(interval=interval, stretch=stretch)
    rgb = map_.make_rgb_image(image_r, image_g, image_b, output_dtype=output_dtype)

    if filename:
        import matplotlib.image

        matplotlib.image.imsave(filename, rgb, origin="lower")

    return rgb



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module includes helper functions for array operations.
"""

import numpy as np

from .decorators import support_nddata

__all__ = ["block_reduce", "block_replicate", "reshape_as_blocks"]


def _process_block_inputs(data, block_size):
    data = np.asanyarray(data)
    block_size = np.atleast_1d(block_size)

    if np.any(block_size <= 0):
        raise ValueError("block_size elements must be strictly positive")

    if data.ndim > 1 and len(block_size) == 1:
        block_size = np.repeat(block_size, data.ndim)

    if len(block_size) != data.ndim:
        raise ValueError(
            "block_size must be a scalar or have the same "
            "length as the number of data dimensions"
        )

    block_size_int = block_size.astype(int)
    if np.any(block_size_int != block_size):  # e.g., 2.0 is OK, 2.1 is not
        raise ValueError("block_size elements must be integers")

    return data, block_size_int


def reshape_as_blocks(data, block_size):
    """
    Reshape a data array into blocks.

    This is useful to efficiently apply functions on block subsets of
    the data instead of using loops.  The reshaped array is a view of
    the input data array.

    .. versionadded:: 4.1

    Parameters
    ----------
    data : ndarray
        The input data array.

    block_size : int or array-like (int)
        The integer block size along each axis.  If ``block_size`` is a
        scalar and ``data`` has more than one dimension, then
        ``block_size`` will be used for for every axis.  Each dimension
        of ``block_size`` must divide evenly into the corresponding
        dimension of ``data``.

    Returns
    -------
    output : ndarray
        The reshaped array as a view of the input ``data`` array.

    Examples
    --------
    >>> import numpy as np
    >>> from astropy.nddata import reshape_as_blocks
    >>> data = np.arange(16).reshape(4, 4)
    >>> data
    array([[ 0,  1,  2,  3],
           [ 4,  5,  6,  7],
           [ 8,  9, 10, 11],
           [12, 13, 14, 15]])
    >>> reshape_as_blocks(data, (2, 2))
    array([[[[ 0,  1],
             [ 4,  5]],
            [[ 2,  3],
             [ 6,  7]]],
           [[[ 8,  9],
             [12, 13]],
            [[10, 11],
             [14, 15]]]])
    """
    data, block_size = _process_block_inputs(data, block_size)

    if np.any(np.mod(data.shape, block_size) != 0):
        raise ValueError(
            "Each dimension of block_size must divide evenly "
            "into the corresponding dimension of data"
        )

    nblocks = np.array(data.shape) // block_size
    new_shape = tuple(k for ij in zip(nblocks, block_size) for k in ij)
    nblocks_idx = tuple(range(0, len(new_shape), 2))  # even indices
    block_idx = tuple(range(1, len(new_shape), 2))  # odd indices

    return data.reshape(new_shape).transpose(nblocks_idx + block_idx)


@support_nddata
def block_reduce(data, block_size, func=np.sum):
    """
    Downsample a data array by applying a function to local blocks.

    If ``data`` is not perfectly divisible by ``block_size`` along a
    given axis then the data will be trimmed (from the end) along that
    axis.

    Parameters
    ----------
    data : array-like
        The data to be resampled.

    block_size : int or array-like (int)
        The integer block size along each axis.  If ``block_size`` is a
        scalar and ``data`` has more than one dimension, then
        ``block_size`` will be used for for every axis.

    func : callable, optional
        The method to use to downsample the data. Must be a callable
        that takes in a 4D `~numpy.ndarray` (the 2D `~numpy.ndarray`
        input into `block_reduce` gets reshaped as 4D) and has an
        ``axis`` keyword that accepts tuples. This function will be
        called with ``axis=(2, 3)`` and it should return a 2D array. The
        default is `~numpy.sum`, which provides block summation (and
        conserves the data sum).

    Returns
    -------
    output : array-like
        The resampled data. Note the depending on the input ``func``,
        the dtype of the output array may not match the input array.

    Examples
    --------
    >>> import numpy as np
    >>> from astropy.nddata import block_reduce
    >>> data = np.arange(16).reshape(4, 4)
    >>> block_reduce(data, 2)  # doctest: +FLOAT_CMP
    array([[10, 18],
           [42, 50]])

    >>> block_reduce(data, 2, func=np.mean)  # doctest: +FLOAT_CMP
    array([[  2.5,   4.5],
           [ 10.5,  12.5]])
    """
    data, block_size = _process_block_inputs(data, block_size)
    nblocks = np.array(data.shape) // block_size
    size_init = nblocks * block_size  # evenly-divisible size

    # trim data if necessary
    for axis in range(data.ndim):
        if data.shape[axis] != size_init[axis]:
            data = data.swapaxes(0, axis)
            data = data[: size_init[axis]]
            data = data.swapaxes(0, axis)

    reshaped = reshape_as_blocks(data, block_size)
    axis = tuple(range(data.ndim, reshaped.ndim))

    return func(reshaped, axis=axis)


@support_nddata
def block_replicate(data, block_size, conserve_sum=True):
    """
    Upsample a data array by block replication.

    Parameters
    ----------
    data : array-like
        The data to be block replicated.

    block_size : int or array-like (int)
        The integer block size along each axis.  If ``block_size`` is a
        scalar and ``data`` has more than one dimension, then
        ``block_size`` will be used for for every axis.

    conserve_sum : bool, optional
        If `True` (the default) then the sum of the output
        block-replicated data will equal the sum of the input ``data``.

    Returns
    -------
    output : array-like
        The block-replicated data. Note that when ``conserve_sum`` is
        `True`, the dtype of the output array will be float.

    Examples
    --------
    >>> import numpy as np
    >>> from astropy.nddata import block_replicate
    >>> data = np.array([[0., 1.], [2., 3.]])
    >>> block_replicate(data, 2)  # doctest: +FLOAT_CMP
    array([[0.  , 0.  , 0.25, 0.25],
           [0.  , 0.  , 0.25, 0.25],
           [0.5 , 0.5 , 0.75, 0.75],
           [0.5 , 0.5 , 0.75, 0.75]])

    >>> block_replicate(data, 2, conserve_sum=False)  # doctest: +FLOAT_CMP
    array([[0., 0., 1., 1.],
           [0., 0., 1., 1.],
           [2., 2., 3., 3.],
           [2., 2., 3., 3.]])
    """
    data, block_size = _process_block_inputs(data, block_size)
    for i in range(data.ndim):
        data = np.repeat(data, block_size[i], axis=i)

    if conserve_sum:
        # in-place division can fail due to dtype casting rule
        data = data / np.prod(block_size)

    return data



# Licensed under a 3-clause BSD style license - see LICENSE.rst

__all__ = ["FLRW", "FlatFLRWMixin"]

import inspect
import warnings
from dataclasses import field
from functools import cached_property
from inspect import signature
from math import floor, pi, sqrt
from numbers import Number
from typing import Any, Final, NamedTuple, TypeVar, overload

import numpy as np
from numpy import inf, sin
from numpy.typing import ArrayLike, NDArray

import astropy.constants as const
import astropy.units as u
from astropy.cosmology._src.typing import CosmoMeta, FArray
from astropy.utils.decorators import lazyproperty
from astropy.utils.exceptions import AstropyUserWarning

# isort: split
from astropy.cosmology._src.core import (
    Cosmology,
    FlatCosmologyMixin,
    dataclass_decorator,
)
from astropy.cosmology._src.parameter import (
    Parameter,
    validate_non_negative,
    validate_with_unit,
)
from astropy.cosmology._src.scipy_compat import quad
from astropy.cosmology._src.traits import (
    BaryonComponent,
    CriticalDensity,
    CurvatureComponent,
    DarkEnergyComponent,
    DarkMatterComponent,
    HubbleParameter,
    MatterComponent,
    PhotonComponent,
    ScaleFactor,
    TemperatureCMB,
    TotalComponent,
)
from astropy.cosmology._src.utils import (
    aszarr,
    deprecated_keywords,
    vectorize_redshift_method,
)

__doctest_requires__ = {"*": ["scipy"]}
_InputT = TypeVar("_InputT", bound=u.Quantity | np.ndarray | np.generic | Number)


##############################################################################
# Parameters

# Some conversion constants -- useful to compute them once here and reuse in
# the initialization rather than have every object do them.

# angle conversions
RAD_IN_ARCSEC: Final = (1 * u.rad).to(u.arcsec)
RAD_IN_ARCMIN: Final = (1 * u.rad).to(u.arcmin)
# Radiation parameter over c^2 in cgs (g cm^-3 K^-4)
a_B_c2: Final = (4 * const.sigma_sb / const.c**3).cgs.value
# Boltzmann constant in eV / K
kB_evK: Final = const.k_B.to(u.eV / u.K)

# See Komatsu et al. 2011, eq 26 and the surrounding discussion for an explanation of
# what this does. However, this is modified to handle multiple neutrino masses by
# computing the above for each mass, then summing
NEUTRINO_FERMI_DIRAC_CORRECTION: Final = 0.22710731766  # 7/8 (4/11)^4/3

# These are purely fitting constants -- see the Komatsu paper
KOMATSU_P: Final = 1.83
KOMATSU_INVP: Final = 0.54644808743  # 1.0 / p
KOMATSU_K: Final = 0.3173

# typing
_FLRWT = TypeVar("_FLRWT", bound="FLRW")
_FlatFLRWMixinT = TypeVar("_FlatFLRWMixinT", bound="FlatFLRWMixin")


##############################################################################


class NeutrinoInfo(NamedTuple):
    """A container for neutrino information.

    This is Private API.

    """

    n_nu: int
    """Number of neutrino species (floor of Neff)."""

    neff_per_nu: float | None
    """Number of effective neutrino species per neutrino.

    We are going to share Neff between the neutrinos equally. In detail this is not
    correct, but it is a standard assumption because properly calculating it is a)
    complicated b) depends on the details of the massive neutrinos (e.g., their weak
    interactions, which could be unusual if one is considering sterile neutrinos).
    """

    has_massive_nu: bool
    """Boolean of which neutrinos are massive."""

    n_massive_nu: int
    """Number of massive neutrinos."""

    n_massless_nu: int
    """Number of massless neutrinos."""

    nu_y: NDArray[np.floating] | None
    """The ratio m_nu / (kB T_nu) for each massive neutrino."""

    nu_y_list: list[float] | None
    """The ratio m_nu / (kB T_nu) for each massive neutrino as a list."""


##############################################################################


ParameterOde0 = Parameter(
    doc="Omega dark energy; dark energy density/critical density at z=0.",
    fvalidate="float",
)


@dataclass_decorator
class FLRW(
    Cosmology,
    # Traits
    BaryonComponent,
    TotalComponent,
    CriticalDensity,
    MatterComponent,
    CurvatureComponent,
    DarkEnergyComponent,
    HubbleParameter,
    PhotonComponent,
    ScaleFactor,
    DarkMatterComponent,
    TemperatureCMB,
):
    """An isotropic and homogeneous (Friedmann-Lemaitre-Robertson-Walker) cosmology.

    This is an abstract base class -- you cannot instantiate examples of this
    class, but must work with one of its subclasses, such as
    :class:`~astropy.cosmology.LambdaCDM` or :class:`~astropy.cosmology.wCDM`.

    Parameters
    ----------
    H0 : float or scalar quantity-like ['frequency']
        Hubble constant at z = 0.  If a float, must be in [km/sec/Mpc].

    Om0 : float
        Omega matter: density of non-relativistic matter in units of the
        critical density at z=0. Note that this does not include massive
        neutrinos.

    Ode0 : float
        Omega dark energy: density of dark energy in units of the critical
        density at z=0.

    Tcmb0 : float or scalar quantity-like ['temperature'], optional
        Temperature of the CMB z=0. If a float, must be in [K]. Default: 0 [K].
        Setting this to zero will turn off both photons and neutrinos
        (even massive ones).

    Neff : float, optional
        Effective number of Neutrino species. Default 3.04.

    m_nu : quantity-like ['energy', 'mass'] or array-like, optional
        Mass of each neutrino species in [eV] (mass-energy equivalency enabled).
        If this is a scalar Quantity, then all neutrino species are assumed to
        have that mass. Otherwise, the mass of each species. The actual number
        of neutrino species (and hence the number of elements of m_nu if it is
        not scalar) must be the floor of Neff. Typically this means you should
        provide three neutrino masses unless you are considering something like
        a sterile neutrino.

    Ob0 : float, optional
        Omega baryons: density of baryonic matter in units of the critical
        density at z=0.

    name : str or None (optional, keyword-only)
        Name for this cosmological object.

    meta : mapping or None (optional, keyword-only)
        Metadata for the cosmology, e.g., a reference.

    Notes
    -----
    Class instances are immutable -- you cannot change the parameters' values.
    That is, all of the above attributes (except meta) are read only.

    For details on how to create performant custom subclasses, see the
    documentation on :ref:`astropy-cosmology-fast-integrals`.
    """

    H0: Parameter = Parameter(
        doc="Hubble constant at z=0.",
        unit="km/(s Mpc)",
        fvalidate="scalar",
    )
    Om0: Parameter = Parameter(
        doc="Omega matter; matter density/critical density at z=0.",
        fvalidate="non-negative",
    )
    Ode0: Parameter = ParameterOde0.clone()
    Tcmb0: Parameter = Parameter(
        default=0.0 * u.K,
        doc="Temperature of the CMB at z=0.",
        unit="Kelvin",
        fvalidate="scalar",
    )
    Neff: Parameter = Parameter(
        default=3.04,
        doc="Number of effective neutrino species.",
        fvalidate="non-negative",
    )
    m_nu: Parameter = Parameter(
        default=0.0 * u.eV,
        doc="Mass of neutrino species.",
        unit="eV",
        equivalencies=u.mass_energy(),
    )
    Ob0: Parameter = Parameter(
        default=0.0,
        doc="Omega baryon; baryonic matter density/critical density at z=0.",
    )

    def __post_init__(self) -> None:
        # Compute neutrino parameters:
        if self.m_nu is None:
            nu_info = NeutrinoInfo(
                n_nu=0,
                neff_per_nu=None,
                has_massive_nu=False,
                n_massive_nu=0,
                n_massless_nu=0,
                nu_y=None,
                nu_y_list=None,
            )
        else:
            n_nu = floor(self.Neff)
            massive = np.nonzero(self.m_nu.value > 0)[0]
            has_massive_nu = massive.size > 0
            n_massive_nu = len(massive)

            # Compute Neutrino Omega and total relativistic component for massive
            # neutrinos. We also store a list version, since that is more efficient
            # to do integrals with (perhaps surprisingly! But small python lists
            # are more efficient than small NumPy arrays).
            if has_massive_nu:
                nu_y = (self.m_nu[massive] / (kB_evK * self.Tnu0)).to_value(u.one)
                nu_y_list = nu_y.tolist()
            else:
                nu_y = nu_y_list = None

            nu_info = NeutrinoInfo(
                n_nu=n_nu,
                # We share Neff between the neutrinos equally. In detail this is not
                # correct. See NeutrinoInfo for more info.
                neff_per_nu=self.Neff / n_nu,
                # Now figure out if we have massive neutrinos to deal with, and if
                # so, get the right number of masses. It is worth keeping track of
                # massless ones separately (since they are easy to deal with, and a
                # common use case is to have only one massive neutrino).
                has_massive_nu=has_massive_nu,
                n_massive_nu=n_massive_nu,
                n_massless_nu=n_nu - n_massive_nu,
                nu_y=nu_y,
                nu_y_list=nu_y_list,
            )

        self._nu_info: NeutrinoInfo
        object.__setattr__(self, "_nu_info", nu_info)

        # Subclasses should override this reference if they provide
        #  more efficient scalar versions of inv_efunc.
        object.__setattr__(self, "_inv_efunc_scalar", self.inv_efunc)
        object.__setattr__(self, "_inv_efunc_scalar_args", ())

    # ---------------------------------------------------------------
    # Parameter details

    @m_nu.validator
    def m_nu(self, param: Parameter, value: Any) -> FArray | None:
        """Validate neutrino masses to right value, units, and shape.

        There are no neutrinos if floor(Neff) or Tcmb0 are 0. The number of
        neutrinos must match floor(Neff). Neutrino masses cannot be
        negative.
        """
        # Check if there are any neutrinos
        if (n_nu := floor(self.Neff)) == 0 or self.Tcmb0.value == 0:
            return None  # None, regardless of input

        # Validate / set units
        value = validate_with_unit(self, param, value)

        # Check values and data shapes
        if value.shape not in ((), (n_nu,)):
            raise ValueError(
                "unexpected number of neutrino masses — "
                f"expected {n_nu}, got {len(value)}."
            )
        elif np.any(value.value < 0):
            raise ValueError("invalid (negative) neutrino mass encountered.")

        # scalar -> array
        if value.isscalar:
            value = np.full_like(value, value, shape=n_nu)

        return value

    # ---------------------------------------------------------------
    # Baryons

    @Ob0.validator
    def Ob0(self, param: Parameter, value: Any) -> float:
        """Validate baryon density to a non-negative float > matter density."""
        if value is None:
            warnings.warn(
                "Ob0=None is deprecated, use Ob0=0 instead, "
                "which never causes methods to raise exceptions.",
                category=DeprecationWarning,
                stacklevel=2,
            )
            return 0.0

        value = validate_non_negative(self, param, value)
        if value > self.Om0:
            raise ValueError(
                "baryonic density can not be larger than total matter density."
            )
        return value

    # ---------------------------------------------------------------
    # Critical Density

    @cached_property
    def critical_density0(self) -> u.Quantity:
        r"""Critical mass density at z=0.

        The critical density is the density of the Universe at which the Universe is
        flat. It is defined as :math:`\rho_{\text{crit}} = 3 H_0^2 / (8 \pi G)`.

        """
        return (3 * self.H0**2 / (8 * pi * const.G)).cgs

    # ---------------------------------------------------------------
    # Curvature

    @cached_property
    def Ok0(self) -> float | np.floating:
        """Omega curvature; the effective curvature density/critical density at z=0."""
        return 1.0 - self.Om0 - self.Ode0 - self.Ogamma0 - self.Onu0

    @property
    def is_flat(self) -> bool:
        """Return `bool`; `True` if the cosmology is globally flat."""
        return bool((self.Ok0 == 0.0) and (self.Otot0 == 1.0))

    # ---------------------------------------------------------------
    # Dark Matter

    @cached_property
    def Odm0(self) -> float:
        """Omega dark matter; dark matter density/critical density at z=0."""
        return self.Om0 - self.Ob0

    # ---------------------------------------------------------------
    # Hubble Parameter

    @deprecated_keywords("z", since="7.0")
    def efunc(self, z: u.Quantity | ArrayLike) -> FArray:
        """Function used to calculate H(z), the Hubble parameter.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        E : array
            The redshift scaling of the Hubble constant.
            Defined such that :math:`H(z) = H_0 E(z)`.

        Notes
        -----
        It is not necessary to override this method, but if de_density_scale
        takes a particularly simple form, it may be advantageous to.
        """
        Or = self.Ogamma0 + (
            self.Onu0
            if not self._nu_info.has_massive_nu
            else self.Ogamma0 * self.nu_relative_density(z)
        )
        zp1 = aszarr(z) + 1.0  # (converts z [unit] -> z [dimensionless])

        return np.sqrt(
            zp1**2 * ((Or * zp1 + self.Om0) * zp1 + self.Ok0)
            + self.Ode0 * self.de_density_scale(z)
        )

    @deprecated_keywords("z", since="7.0")
    def inv_efunc(self, z: u.Quantity | ArrayLike) -> FArray:
        """Inverse of ``efunc``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        E : array
            The redshift scaling of the inverse Hubble constant.
        """
        # Avoid the function overhead by repeating code
        Or = self.Ogamma0 + (
            self.Onu0
            if not self._nu_info.has_massive_nu
            else self.Ogamma0 * self.nu_relative_density(z)
        )
        zp1 = aszarr(z) + 1.0  # (converts z [unit] -> z [dimensionless])

        return (
            zp1**2 * ((Or * zp1 + self.Om0) * zp1 + self.Ok0)
            + self.Ode0 * self.de_density_scale(z)
        ) ** (-0.5)

    # ---------------------------------------------------------------
    # properties

    @property
    def Otot0(self) -> float:
        """Omega total; the total density/critical density at z=0."""
        return self.Om0 + self.Ogamma0 + self.Onu0 + self.Ode0 + self.Ok0

    @cached_property
    def Tnu0(self) -> u.Quantity:
        """Temperature of the neutrino background as |Quantity| at z=0."""
        # The constant in front is (4/11)^1/3 -- see any cosmology book for an
        # explanation -- for example, Weinberg 'Cosmology' p 154 eq (3.1.21).
        return 0.7137658555036082 * self.Tcmb0

    @property
    def has_massive_nu(self) -> bool:
        """Does this cosmology have at least one massive neutrino species?"""
        if self.Tnu0.value == 0:
            return False
        return self._nu_info.has_massive_nu

    @cached_property
    def Ogamma0(self) -> float:
        """Omega gamma; the density/critical density of photons at z=0."""
        # photon density from Tcmb
        return a_B_c2 * self.Tcmb0.value**4 / self.critical_density0.value

    @cached_property
    def Onu0(self) -> float:
        """Omega nu; the density/critical density of neutrinos at z=0."""
        if self._nu_info.has_massive_nu:
            return self.Ogamma0 * self.nu_relative_density(0)
        else:
            # This case is particularly simple, so do it directly The 0.2271...
            # is 7/8 (4/11)^(4/3) -- the temperature bit ^4 (blackbody energy
            # density) times 7/8 for FD vs. BE statistics.
            return NEUTRINO_FERMI_DIRAC_CORRECTION * self.Neff * self.Ogamma0

    # ---------------------------------------------------------------

    @deprecated_keywords("z", since="7.0")
    def Otot(self, z: u.Quantity | ArrayLike) -> FArray:
        """The total density parameter at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshifts.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        Otot : array
            The total density relative to the critical density at each redshift.
        """
        return self.Om(z) + self.Ogamma(z) + self.Onu(z) + self.Ode(z) + self.Ok(z)

    # Odm is provided by the DarkMatterComponent trait
    # Ogamma is provided by the PhotonComponent trait

    @deprecated_keywords("z", since="7.0")
    def Onu(self, z: u.Quantity | ArrayLike) -> FArray:
        r"""Return the density parameter for neutrinos at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        Onu : ndarray
            The energy density of neutrinos relative to the critical density at
            each redshift. Note that this includes their kinetic energy (if
            they have mass), so it is not equal to the commonly used
            :math:`\sum \frac{m_{\nu}}{94 eV}`, which does not include
            kinetic energy.
        """
        z = aszarr(z)
        if self.Onu0 == 0:  # Common enough to be worth checking explicitly
            return np.zeros_like(z)
        return self.Ogamma(z) * self.nu_relative_density(z)

    @deprecated_keywords("z", since="7.0")
    def Tnu(self, z: u.Quantity | ArrayLike) -> u.Quantity:
        """Return the neutrino temperature at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        Tnu : Quantity ['temperature']
            The temperature of the cosmic neutrino background in K.
        """
        return self.Tnu0 * (aszarr(z) + 1.0)

    @deprecated_keywords("z", since="7.0")
    def nu_relative_density(self, z: u.Quantity | ArrayLike) -> FArray:
        r"""Neutrino density function relative to the energy density in photons.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        f : array
            The neutrino density scaling factor relative to the density in
            photons at each redshift.

        Notes
        -----
        The density in neutrinos is given by

        .. math::

           \rho_{\nu} \left(a\right) = 0.2271 \, N_{eff} \,
           f\left(m_{\nu} a / T_{\nu 0} \right) \,
           \rho_{\gamma} \left( a \right)

        where

        .. math::

           f \left(y\right) = \frac{120}{7 \pi^4}
           \int_0^{\infty} \, dx \frac{x^2 \sqrt{x^2 + y^2}}
           {e^x + 1}

        assuming that all neutrino species have the same mass.
        If they have different masses, a similar term is calculated for each
        one. Note that ``f`` has the asymptotic behavior :math:`f(0) = 1`. This
        method returns :math:`0.2271 f` using an analytical fitting formula
        given in Komatsu et al. 2011, ApJS 192, 18.
        """
        # Note that there is also a scalar-z-only cython implementation of
        # this in scalar_inv_efuncs.pyx, so if you find a problem in this
        # you need to update there too.

        # The massive and massless contribution must be handled separately
        # But check for common cases first
        z = aszarr(z)
        if not self._nu_info.has_massive_nu:
            return NEUTRINO_FERMI_DIRAC_CORRECTION * self.Neff * np.ones_like(z)

        curr_nu_y = self._nu_info.nu_y / (1.0 + np.expand_dims(z, axis=-1))
        rel_mass_per = (1.0 + (KOMATSU_K * curr_nu_y) ** KOMATSU_P) ** KOMATSU_INVP
        rel_mass = rel_mass_per.sum(-1) + self._nu_info.n_massless_nu

        return NEUTRINO_FERMI_DIRAC_CORRECTION * self._nu_info.neff_per_nu * rel_mass

    def _lookback_time_integrand_scalar(self, z: float, /) -> float:
        """Integrand of the lookback time (equation 30 of [1]_).

        Parameters
        ----------
        z : float, positional-only
            Input redshift.

            .. versionchanged:: 7.0
                The argument is positional-only.

        Returns
        -------
        I : float
            The integrand for the lookback time.

        References
        ----------
        .. [1] Hogg, D. (1999). Distance measures in cosmology, section 11.
               arXiv e-prints, astro-ph/9905116.
        """
        return self._inv_efunc_scalar(z, *self._inv_efunc_scalar_args) / (z + 1.0)

    @deprecated_keywords("z", since="7.0")
    def lookback_time_integrand(self, z: u.Quantity | ArrayLike) -> FArray:
        """Integrand of the lookback time (equation 30 of [1]_).

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        I : array
            The integrand for the lookback time.

        References
        ----------
        .. [1] Hogg, D. (1999). Distance measures in cosmology, section 11.
               arXiv e-prints, astro-ph/9905116.
        """
        z = aszarr(z)
        return self.inv_efunc(z) / (z + 1.0)

    def _abs_distance_integrand_scalar(self, z: u.Quantity | ArrayLike, /) -> float:
        """Integrand of the absorption distance (eq. 4, [1]_).

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like, positional-only
            Input redshift.

            .. versionchanged:: 7.0
                The argument is positional-only.

        Returns
        -------
        dX : float
            The integrand for the absorption distance (dimensionless).

        References
        ----------
        .. [1] Bahcall, John N. and Peebles, P.J.E. 1969, ApJ, 156L, 7B
        """
        return (z + 1.0) ** 2 * self._inv_efunc_scalar(z, *self._inv_efunc_scalar_args)

    @deprecated_keywords("z", since="7.0")
    def abs_distance_integrand(self, z: u.Quantity | ArrayLike) -> FArray:
        """Integrand of the absorption distance (eq. 4, [1]_).

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        dX : array
            The integrand for the absorption distance (dimensionless).

        References
        ----------
        .. [1] Bahcall, John N. and Peebles, P.J.E. 1969, ApJ, 156L, 7B
        """
        z = aszarr(z)
        return (z + 1.0) ** 2 * self.inv_efunc(z)

    @deprecated_keywords("z", since="7.0")
    def lookback_time(self, z: u.Quantity | ArrayLike) -> u.Quantity:
        """Lookback time in Gyr to redshift ``z``.

        The lookback time is the difference between the age of the Universe now
        and the age at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        t : Quantity ['time']
            Lookback time in Gyr to each input redshift.

        See Also
        --------
        z_at_value : Find the redshift corresponding to a lookback time.
        """
        return self._lookback_time(z)

    def _lookback_time(self, z: u.Quantity | ArrayLike, /) -> u.Quantity:
        """Lookback time in Gyr to redshift ``z``.

        The lookback time is the difference between the age of the Universe now
        and the age at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like, positional-only
            Input redshift.

            .. versionchanged:: 7.0
                The argument is positional-only.

        Returns
        -------
        t : Quantity ['time']
            Lookback time in Gyr to each input redshift.
        """
        return self.hubble_time * self._integral_lookback_time(z)

    @vectorize_redshift_method
    def _integral_lookback_time(self, z: u.Quantity | ArrayLike, /) -> FArray:
        """Lookback time to redshift ``z``. Value in units of Hubble time.

        The lookback time is the difference between the age of the Universe now
        and the age at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like, positional-only
            Input redshift.

            .. versionchanged:: 7.0
                The argument is positional-only.

        Returns
        -------
        t : ndarray
            Lookback time to each input redshift in Hubble time units.
        """
        return quad(self._lookback_time_integrand_scalar, 0, z)[0]

    @deprecated_keywords("z", since="7.0")
    def lookback_distance(self, z: u.Quantity | ArrayLike) -> u.Quantity:
        """The lookback distance is the light travel time distance to a given redshift.

        It is simply c * lookback_time. It may be used to calculate
        the proper distance between two redshifts, e.g. for the mean free path
        to ionizing radiation.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        d : Quantity ['length']
            Lookback distance in Mpc
        """
        return (self.lookback_time(z) * const.c).to(u.Mpc)

    @deprecated_keywords("z", since="7.0")
    def age(self, z: u.Quantity | ArrayLike) -> u.Quantity:
        """Age of the universe in Gyr at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        t : Quantity ['time']
            The age of the universe in Gyr at each input redshift.

        See Also
        --------
        z_at_value : Find the redshift corresponding to an age.
        """
        return self._age(z)

    def _age(self, z: u.Quantity | ArrayLike, /) -> u.Quantity:
        """Age of the universe in Gyr at redshift ``z``.

        This internal function exists to be re-defined for optimizations.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like, positional-only
            Input redshift.

            .. versionchanged:: 7.0
                The argument is positional-only.

        Returns
        -------
        t : Quantity ['time']
            The age of the universe in Gyr at each input redshift.
        """
        return self.hubble_time * self._integral_age(z)

    @vectorize_redshift_method
    def _integral_age(self, z: u.Quantity | ArrayLike, /) -> FArray:
        """Age of the universe at redshift ``z``. Value in units of Hubble time.

        Calculated using explicit integration.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like, positional-only
            Input redshift.

        Returns
        -------
        t : array
            The age of the universe at each input redshift in Hubble time units.

        See Also
        --------
        z_at_value : Find the redshift corresponding to an age.
        """
        return quad(self._lookback_time_integrand_scalar, z, inf)[0]

    # ---------------------------------------------------------------
    # Comoving distance

    @overload
    def comoving_distance(self, z: _InputT) -> u.Quantity: ...

    @overload
    def comoving_distance(self, z: _InputT, z2: _InputT) -> u.Quantity: ...

    @deprecated_keywords("z2", since="7.1")
    @deprecated_keywords("z", since="7.0")
    def comoving_distance(self, z: _InputT, z2: _InputT | None = None) -> u.Quantity:
        r"""Comoving line-of-sight distance :math:`d_c(z1, z2)` in Mpc.

        The comoving distance along the line-of-sight between two objects
        remains constant with time for objects in the Hubble flow.

        Parameters
        ----------
        z, z2 : Quantity ['redshift'], positional-only
            Input redshifts. If one argument ``z`` is given, the distance
            :math:`d_c(0, z)` is returned. If two arguments ``z1, z2`` are
            given, the distance :math:`d_c(z_1, z_2)` is returned.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        Quantity ['length']
            Comoving distance in Mpc between each input redshift.
        """
        z1, z2 = (0.0, z) if z2 is None else (z, z2)
        return self._comoving_distance_z1z2(z1, z2)

    def _comoving_distance_z1z2(
        self, z1: u.Quantity | ArrayLike, z2: u.Quantity | ArrayLike, /
    ) -> u.Quantity:
        """Comoving line-of-sight distance in Mpc between redshifts ``z1`` and ``z2``.

        The comoving distance along the line-of-sight between two objects
        remains constant with time for objects in the Hubble flow.

        Parameters
        ----------
        z1, z2 : Quantity-like ['redshift'], array-like, positional-only
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        d : Quantity ['length']
            Comoving distance in Mpc between each input redshift.
        """
        return self._integral_comoving_distance_z1z2(z1, z2)

    def _integral_comoving_distance_z1z2(
        self, z1: u.Quantity | ArrayLike, z2: u.Quantity | ArrayLike, /
    ) -> u.Quantity:
        """Comoving line-of-sight distance (Mpc) between objects at redshifts z1 and z2.

        The comoving distance along the line-of-sight between two objects remains
        constant with time for objects in the Hubble flow.

        Parameters
        ----------
        z1, z2 : Quantity-like ['redshift'] or array-like
            Input redshifts.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        |Quantity| ['length']
            Comoving distance in Mpc between each input redshift.
        """
        return self.hubble_distance * self._integral_comoving_distance_z1z2_scalar(z1, z2)  # fmt: skip

    @vectorize_redshift_method(nin=2)
    def _integral_comoving_distance_z1z2_scalar(
        self, z1: u.Quantity | ArrayLike, z2: u.Quantity | ArrayLike, /
    ) -> FArray:
        """Comoving line-of-sight distance in Mpc between objects at redshifts ``z1`` and ``z2``.

        The comoving distance along the line-of-sight between two objects
        remains constant with time for objects in the Hubble flow.

        Parameters
        ----------
        z1, z2 : Quantity-like ['redshift'], array-like, positional-only
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        d : array
            Comoving distance in Mpc between each input redshift.
        """
        return quad(self._inv_efunc_scalar, z1, z2, args=self._inv_efunc_scalar_args)[0]

    # ---------------------------------------------------------------

    @deprecated_keywords("z", since="7.0")
    def comoving_transverse_distance(self, z: u.Quantity | ArrayLike) -> u.Quantity:
        r"""Comoving transverse distance in Mpc at a given redshift.

        This value is the transverse comoving distance at redshift ``z``
        corresponding to an angular separation of 1 radian. This is the same as
        the comoving distance if :math:`\Omega_k` is zero (as in the current
        concordance Lambda-CDM model).

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        d : Quantity ['length']
            Comoving transverse distance in Mpc at each input redshift.

        Notes
        -----
        This quantity is also called the 'proper motion distance' in some texts.
        """
        return self._comoving_transverse_distance_z1z2(0, z)

    def _comoving_transverse_distance_z1z2(
        self, z1: u.Quantity | ArrayLike, z2: u.Quantity | ArrayLike, /
    ) -> u.Quantity:
        r"""Comoving transverse distance in Mpc between two redshifts.

        This value is the transverse comoving distance at redshift ``z2`` as
        seen from redshift ``z1`` corresponding to an angular separation of
        1 radian. This is the same as the comoving distance if :math:`\Omega_k`
        is zero (as in the current concordance Lambda-CDM model).

        Parameters
        ----------
        z1, z2 : Quantity-like ['redshift'], array-like, positional-only
            Input redshifts.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        d : Quantity ['length']
            Comoving transverse distance in Mpc between input redshift.

        Notes
        -----
        This quantity is also called the 'proper motion distance' in some texts.
        """
        Ok0 = self.Ok0
        dc = self._comoving_distance_z1z2(z1, z2)
        if Ok0 == 0:
            return dc
        sqrtOk0 = sqrt(abs(Ok0))
        dh = self.hubble_distance
        if Ok0 > 0:
            return dh / sqrtOk0 * np.sinh(sqrtOk0 * dc.value / dh.value)
        else:
            return dh / sqrtOk0 * sin(sqrtOk0 * dc.value / dh.value)

    @deprecated_keywords("z", since="7.0")
    def angular_diameter_distance(self, z: u.Quantity | ArrayLike) -> u.Quantity:
        """Angular diameter distance in Mpc at a given redshift.

        This gives the proper (sometimes called 'physical') transverse
        distance corresponding to an angle of 1 radian for an object
        at redshift ``z`` ([1]_, [2]_, [3]_).

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        d : Quantity ['length']
            Angular diameter distance in Mpc at each input redshift.

        References
        ----------
        .. [1] Weinberg, 1972, pp 420-424; Weedman, 1986, pp 421-424.
        .. [2] Weedman, D. (1986). Quasar astronomy, pp 65-67.
        .. [3] Peebles, P. (1993). Principles of Physical Cosmology, pp 325-327.
        """
        z = aszarr(z)
        return self.comoving_transverse_distance(z) / (z + 1.0)

    @deprecated_keywords("z", since="7.0")
    def luminosity_distance(self, z: u.Quantity | ArrayLike) -> u.Quantity:
        """Luminosity distance in Mpc at redshift ``z``.

        This is the distance to use when converting between the bolometric flux
        from an object at redshift ``z`` and its bolometric luminosity [1]_.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        d : Quantity ['length']
            Luminosity distance in Mpc at each input redshift.

        See Also
        --------
        z_at_value : Find the redshift corresponding to a luminosity distance.

        References
        ----------
        .. [1] Weinberg, 1972, pp 420-424; Weedman, 1986, pp 60-62.
        """
        z = aszarr(z)
        return (z + 1.0) * self.comoving_transverse_distance(z)

    def angular_diameter_distance_z1z2(
        self, z1: u.Quantity | ArrayLike, z2: u.Quantity | ArrayLike
    ) -> u.Quantity:
        """Angular diameter distance between objects at 2 redshifts.

        Useful for gravitational lensing, for example computing the angular
        diameter distance between a lensed galaxy and the foreground lens.

        Parameters
        ----------
        z1, z2 : Quantity-like ['redshift'], array-like
            Input redshifts. For most practical applications such as
            gravitational lensing, ``z2`` should be larger than ``z1``. The
            method will work for ``z2 < z1``; however, this will return
            negative distances.

        Returns
        -------
        d : Quantity ['length']
            The angular diameter distance between each input redshift pair.
            Returns scalar if input is scalar, array else-wise.
        """
        z1, z2 = aszarr(z1), aszarr(z2)
        if np.any(z2 < z1):
            warnings.warn(
                f"Second redshift(s) z2 ({z2}) is less than first "
                f"redshift(s) z1 ({z1}).",
                AstropyUserWarning,
            )
        return self._comoving_transverse_distance_z1z2(z1, z2) / (z2 + 1.0)

    @vectorize_redshift_method
    def absorption_distance(self, z: u.Quantity | ArrayLike, /) -> FArray:
        """Absorption distance at redshift ``z`` (eq. 4, [1]_).

        This is used to calculate the number of objects with some cross section
        of absorption and number density intersecting a sightline per unit
        redshift path [1]_.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like, positional-only
            Input redshift.

        Returns
        -------
        X : array
            Absorption distance (dimensionless) at each input redshift.

        References
        ----------
        .. [1] Bahcall, John N. and Peebles, P.J.E. 1969, ApJ, 156L, 7B
        """
        return quad(self._abs_distance_integrand_scalar, 0, z)[0]

    @deprecated_keywords("z", since="7.0")
    def distmod(self, z: u.Quantity | ArrayLike) -> u.Quantity:
        """Distance modulus at redshift ``z``.

        The distance modulus is defined as the (apparent magnitude - absolute
        magnitude) for an object at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        distmod : Quantity ['length']
            Distance modulus at each input redshift, in magnitudes.

        See Also
        --------
        z_at_value : Find the redshift corresponding to a distance modulus.
        """
        # Remember that the luminosity distance is in Mpc
        # Abs is necessary because in certain obscure closed cosmologies
        #  the distance modulus can be negative -- which is okay because
        #  it enters as the square.
        val = 5.0 * np.log10(abs(self.luminosity_distance(z).value)) + 25.0
        return u.Quantity(val, u.mag)

    @deprecated_keywords("z", since="7.0")
    def comoving_volume(self, z: u.Quantity | ArrayLike) -> u.Quantity:
        r"""Comoving volume in cubic Mpc at redshift ``z``.

        This is the volume of the universe encompassed by redshifts less than
        ``z``. For the case of :math:`\Omega_k = 0` it is a sphere of radius
        `comoving_distance` but it is less intuitive if :math:`\Omega_k` is not.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        V : Quantity ['volume']
            Comoving volume in :math:`Mpc^3` at each input redshift.
        """
        Ok0 = self.Ok0
        if Ok0 == 0:
            return 4.0 / 3.0 * pi * self.comoving_distance(z) ** 3

        dh = self.hubble_distance.value  # .value for speed
        dm = self.comoving_transverse_distance(z).value
        term1 = 4.0 * pi * dh**3 / (2.0 * Ok0) * u.Mpc**3
        term2 = dm / dh * np.sqrt(1 + Ok0 * (dm / dh) ** 2)
        term3 = sqrt(abs(Ok0)) * dm / dh

        if Ok0 > 0:
            return term1 * (term2 - 1.0 / sqrt(abs(Ok0)) * np.arcsinh(term3))
        else:
            return term1 * (term2 - 1.0 / sqrt(abs(Ok0)) * np.arcsin(term3))

    @deprecated_keywords("z", since="7.0")
    def differential_comoving_volume(self, z: u.Quantity | ArrayLike) -> u.Quantity:
        """Differential comoving volume at redshift z.

        Useful for calculating the effective comoving volume.
        For example, allows for integration over a comoving volume that has a
        sensitivity function that changes with redshift. The total comoving
        volume is given by integrating ``differential_comoving_volume`` to
        redshift ``z`` and multiplying by a solid angle.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        dV : Quantity
            Differential comoving volume per redshift per steradian at each
            input redshift.
        """
        dm = self.comoving_transverse_distance(z)
        return self.hubble_distance * (dm**2.0) / (self.efunc(z) << u.steradian)

    @deprecated_keywords("z", since="7.0")
    def kpc_comoving_per_arcmin(self, z: u.Quantity | ArrayLike) -> u.Quantity:
        """Separation in transverse comoving kpc equal to an arcmin at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        d : Quantity ['length']
            The distance in comoving kpc corresponding to an arcmin at each
            input redshift.
        """
        return self.comoving_transverse_distance(z).to(u.kpc) / RAD_IN_ARCMIN

    @deprecated_keywords("z", since="7.0")
    def kpc_proper_per_arcmin(self, z: u.Quantity | ArrayLike) -> u.Quantity:
        """Separation in transverse proper kpc equal to an arcminute at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        d : Quantity ['length']
            The distance in proper kpc corresponding to an arcmin at each input
            redshift.
        """
        return self.angular_diameter_distance(z).to(u.kpc) / RAD_IN_ARCMIN

    @deprecated_keywords("z", since="7.0")
    def arcsec_per_kpc_comoving(self, z: u.Quantity | ArrayLike) -> u.Quantity:
        """Angular separation in arcsec equal to a comoving kpc at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        theta : Quantity ['angle']
            The angular separation in arcsec corresponding to a comoving kpc at
            each input redshift.
        """
        return RAD_IN_ARCSEC / self.comoving_transverse_distance(z).to(u.kpc)

    @deprecated_keywords("z", since="7.0")
    def arcsec_per_kpc_proper(self, z: u.Quantity | ArrayLike) -> u.Quantity:
        """Angular separation in arcsec corresponding to a proper kpc at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        theta : Quantity ['angle']
            The angular separation in arcsec corresponding to a proper kpc at
            each input redshift.
        """
        return RAD_IN_ARCSEC / self.angular_diameter_distance(z).to(u.kpc)


@dataclass_decorator
class FlatFLRWMixin(FlatCosmologyMixin):
    """Mixin class for flat FLRW cosmologies.

    Do NOT instantiate directly. Must precede the base class in the
    multiple-inheritance so that this mixin's ``__init__`` proceeds the
    base class'. Note that all instances of ``FlatFLRWMixin`` are flat, but
    not all flat cosmologies are instances of ``FlatFLRWMixin``. As
    example, ``LambdaCDM`` **may** be flat (for the a specific set of
    parameter values), but ``FlatLambdaCDM`` **will** be flat.
    """

    Ode0: Parameter = field(  # now a derived param.
        default=ParameterOde0.clone(default=0, derived=True),
        init=False,
        repr=False,
    )

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        # Check that Ode0 is not in __init__
        if (
            getattr(
                vars(cls).get("Ode0", cls.__dataclass_fields__.get("Ode0")),
                "init",
                True,
            )
            or "Ode0" in signature(cls.__init__).parameters
        ):
            msg = "subclasses of `FlatFLRWMixin` cannot have `Ode0` in `__init__`"
            raise TypeError(msg)

    def __post_init__(self) -> None:
        self.__dict__["Ode0"] = 0
        super().__post_init__()
        # Do some twiddling after the fact to get flatness
        self.__dict__["Ok0"] = 0.0
        self.__dict__["Ode0"] = 1.0 - (self.Om0 + self.Ogamma0 + self.Onu0 + self.Ok0)

    @lazyproperty
    def nonflat(self: _FlatFLRWMixinT) -> _FLRWT:  # noqa: PYI019
        # Create BoundArgument to handle args versus kwargs.
        # This also handles all errors from mismatched arguments
        ba = inspect.signature(self.__nonflatclass__).bind_partial(
            **dict(self.parameters), Ode0=self.Ode0, name=self.name
        )
        # Make new instance, respecting args vs kwargs
        inst = self.__nonflatclass__(*ba.args, **ba.kwargs)
        # Because of machine precision, make sure parameters exactly match
        inst.__dict__.update(inst.parameters)
        inst.__dict__.update(inst._derived_parameters)
        inst.__dict__["Ok0"] = self.Ok0

        return inst

    @property
    def Otot0(self) -> float:
        """Omega total; the total density/critical density at z=0."""
        return 1.0

    @deprecated_keywords("z", since="7.0")
    def Otot(self, z: u.Quantity | ArrayLike) -> FArray:
        """The total density parameter at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        Otot : array
        """
        return np.ones_like(aszarr(z), subok=True)

    def clone(
        self, *, meta: CosmoMeta | None = None, to_nonflat: bool = False, **kwargs: Any
    ) -> "FLRW":
        if not to_nonflat and kwargs.get("Ode0") is not None:
            msg = "Cannot set 'Ode0' in clone unless 'to_nonflat=True'. "
            raise ValueError(msg)
        return super().clone(meta=meta, to_nonflat=to_nonflat, **kwargs)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
The content of this module is solely for internal use of ``astropy``
and subject to changes without deprecations. Do not use it in external
packages or code.

"""

# Importing this module will also install monkey-patches defined in it
from .numpycompat import *
